'''
Created on 12 Jul 2021

@author: jacklok
'''

from wtforms import StringField, PasswordField, validators, DateField
from trexadmin.forms.base_forms import ValidationBaseForm
from trexadmin.libs.wtforms import validators as custom_validator
from flask_babel import gettext
from datetime import date
from wtforms.fields.simple import HiddenField

class CustomerDetailsBaseForm(ValidationBaseForm):
    customer_key        = HiddenField(gettext('Customer Key'), validators=[
                                        ]
                                        )
    name                = StringField(gettext('Name'), validators=[
                                        validators.InputRequired(gettext('Name is required')),
                                        validators.Length(min=3, max=300, message='Name length must be within 3 and 300 characters'),
                                        
                                        ]
                                        )
    gender              = StringField(gettext('Gender'), [
                                        validators.Optional(),
                                        validators.Length(min=1, max=1, message=gettext('Gender value is either m or f')),
                                        
                                        ]
                                        )
    
    birth_date          = DateField('Date of Birth', format='%d/%m/%Y', validators=[
                                            validators.InputRequired(gettext('Birth date is required')),
                                        ])
    
    email               = StringField('Email Address', validators=[
                                        validators.Email(gettext("Please enter valid email address.")),
                                        custom_validator.RequiredIfOtherFieldEmpty(
                                                        ['mobile_phone'],
                                                        message=gettext("Either email or mobile phone is required"),
                                                        
                                                        ),
                                        ]
                                        )
    
    mobile_phone        = StringField('Mobile Phone', validators=[
                                        custom_validator.RequiredIfOtherFieldEmpty(
                                                        ['email'],
                                                        message=gettext("Either email or mobile phone is required"),
                                                        
                                                        ),
                                        
                                        ]
                                        )
    
    merchant_reference_code       = StringField('Member code', validators=[
                                        validators.Optional(),
                                        validators.Length(max=16, message=gettext("Member code length must not more than 16 characters"))
                                        ]
                                        )
    
    
    
class CustomerDetailsNewForm(CustomerDetailsBaseForm):
    
    
    password                      = StringField(gettext('Password'), validators=[
                                        validators.InputRequired(gettext('Password is required')),
                                        
                                        ]
                                        )    

class CustomerDetailsUpdateForm(CustomerDetailsBaseForm):
    
    password                      = StringField(gettext('Password'), validators=[
                                        validators.Optional(),
                                        ]
                                        )  
    
    
class CustomerSearchForm(ValidationBaseForm):
    name                            = StringField('Name', [
                                        validators.Optional(),
                                        validators.Length(min=3, max=300, message='Name length must be within 3 and 300 characters'),
                                        custom_validator.RequiredIfOtherFieldsValueIsEmpty(
                                                        ['mobile_phone', 'email', 'reference_code', 'merchant_reference_code', ],
                                                        message=gettext("Either one input is required")),
                                        ]
                                        )
    email                           = StringField('Email Address', [
                                        validators.Optional(),
                                        validators.Email("Please enter valid email address."),
                                        custom_validator.RequiredIfOtherFieldsValueIsEmpty(
                                                        ['name', 'mobile_phone', 'reference_code', 
                                                         'merchant_reference_code', ],
                                                        message=gettext("Either one input is required")),
                                        
                                        ]
                                        )
    
    mobile_phone                    = StringField('Mobile Phone', [
                                        validators.Optional(),
                                        custom_validator.RequiredIfOtherFieldsValueIsEmpty(
                                                        ['name', 'email', 'reference_code', 'merchant_reference_code', ],
                                                        message=gettext("Either one input is required")),
                                        
                                        ]
                                        )
    
    reference_code                  = StringField('Reference code', [
                                        validators.Optional(),
                                        custom_validator.RequiredIfOtherFieldsValueIsEmpty(
                                                        ['mobile_phone', 'email', 'name', 'merchant_reference_code',  ],
                                                        message=gettext("Either one input is required")),
                                        validators.Length(max=16, message="Reference code length must not more than 16 characters")
                                        ]
                                        )
    
    merchant_reference_code         = StringField('Member code', [
                                        validators.Optional(),
                                        custom_validator.RequiredIfOtherFieldsValueIsEmpty(
                                                        ['mobile_phone', 'email', 'name', 'merchant_reference_code', 
                                                         ],
                                                        message=gettext("Either one input is required")),
                                        validators.Length(max=16, message="Member code length must not more than 16 characters")
                                        ]
                                        )
    
    


'''
Created on 8 Jul 2021

@author: jacklok
'''
from flask import request
import logging
from trexlib.utils.crypto_util import decrypt_json
from datetime import datetime, timedelta
from trexapi import conf as api_conf
from trexlib.utils.crypto_util import encrypt_json

logger = logging.getLogger('helper')


def get_logged_in_api_username():
    auth_token  = request.headers.get('x-auth-token')
    username    = None
    try:
        auth_details_json = decrypt_json(auth_token)
    except:
        logger.error('Failed to decrypt authenticated token')
        
    logger.debug('auth_details_json=%s', auth_details_json)
    
    if auth_details_json:
        username = auth_details_json.get('username')
        
    return username

def generate_user_auth_token(acct_id, reference_code):
    expiry_datetime = datetime.now() + timedelta(minutes = int(api_conf.API_TOKEN_EXPIRY_LENGTH_IN_MINUTE))
    
    logger.debug('expiry_datetime=%s', expiry_datetime)
    
    token_content =  {
                        'acct_id'           : acct_id,
                        'reference_code'    : reference_code,
                        'expiry_datetime'   : expiry_datetime.strftime('%d-%m-%Y %H:%M:%S'),
                        }
    
    logger.debug('token_content=%s', token_content)
    
    return (expiry_datetime, encrypt_json(token_content))



'''
Created on 30 Jun 2021

@author: jacklok
'''
import os
APPLICATION_NAME                                                = os.environ.get('APPLICATION_NAME')
MOBILE_APP_NAME                                                 = os.environ.get('MOBILE_APP_NAME')
APPLICATION_BASE_URL                                            = os.environ.get('APPLICATION_BASE_URL')
SECRET_KEY                                                      = os.environ.get('SECRET_KEY')
VERSION                                                         = '0.0.1'
UPDATED_DATE                                                    = '5 July 2021'

API_TOKEN_EXPIRY_LENGTH_IN_MINUTE                               = os.environ.get('API_TOKEN_EXPIRY_LENGTH_IN_MINUTE') 

API_ERR_CODE_INVALID_ACTIVATION_CODE                            = 'invalid.activation.code';

EARN_INSTANT_REWARD_URL                                         = os.environ.get('EARN_INSTANT_REWARD_URL')

LUCKY_DRAW_URL                                                  = os.environ.get('LUCKY_DRAW_URL')



from flask import Blueprint, request
import logging
from trexmodel.utils.model.model_util import create_db_client
from trexadmin.libs.http import create_rest_message
from trexadmin.libs.http import StatusCode
from trexmodel.models.datastore.merchant_models import MerchantAcct, Outlet,\
    MerchantUser
from trexmodel.models.datastore.customer_models import Customer,\
    CustomerMembership
from trexapi.decorators.api_decorators import user_auth_token_required,\
    auth_token_required
from flask.json import jsonify
from trexadmin.libs.decorators import elapsed_time_trace
from trexmodel.models.datastore.model_decorators import model_transactional
from trexmodel.models.datastore.membership_models import MerchantMembership
from trexapi.utils.api_helpers import get_logged_in_api_username
from trexmodel.models.datastore.transaction_models import CustomerTransaction
from trexapi.utils.reward_transaction_helper import check_giveaway_reward_for_membership_purchase_transaction
from datetime import datetime


customer_membership_api_bp = Blueprint('customer_membership_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/customer-membership')

logger = logging.getLogger('debug')


@customer_membership_api_bp.route('/ping', methods=['GET'])
def ping():
    return create_rest_message('OK', status_code=StatusCode.OK)

@customer_membership_api_bp.route('/customer/reference-code/<reference_code>/assign-membership', methods=['POST'])
@auth_token_required
#@elapsed_time_trace(trace_key="assign_membership")
def assign_membership(reference_code):
    acct_id         = request.headers.get('x-acct-id')
    outlet_key      = request.headers.get('x-outlet-key')
    logger.debug('reference_code=%s', reference_code)
    logger.debug('acct_id=%s', acct_id)
    
    merchant_membership_key = request.args.get('merchant_membership_key') or request.form.get('merchant_membership_key') or request.json.get('merchant_membership_key')
    
    db_client = create_db_client(caller_info="assign_membership")
    customer_membership = None
    customer            = None
    with db_client.context():
        merchant_acct = MerchantAcct.fetch(acct_id)
        customer = Customer.get_by_reference_code(reference_code, merchant_acct)
        if customer:
            merchant_membership = MerchantMembership.fetch(merchant_membership_key)
                        
            customer_membership = CustomerMembership.get_by_customer_and_merchant_membership(customer, merchant_membership)
            
    if customer_membership:
        return create_rest_message('Customer membership already have been assigned', status_code=StatusCode.BAD_REQUEST)
    else:
        with db_client.context():    
            merchant_username       = get_logged_in_api_username()
            assigned_by             = MerchantUser.get_by_username(merchant_username)
            assigned_outlet         = Outlet.fetch(outlet_key)
            if customer:
                __assign_membership(customer, merchant_membership, assigned_by, assigned_outlet)
            
            
    
        return create_rest_message('Customer membership have been assigned successfully', status_code=StatusCode.OK)
        
    
    return create_rest_message(status_code=StatusCode.OK)

@model_transactional(desc="assign customer membership")
def __assign_membership(customer, merchant_membership, assigned_by, assigned_outlet):
    customer_membership = CustomerMembership.create(customer, merchant_membership, assigned_by=assigned_by, assigned_outlet=assigned_outlet)
    
    customer_transaction = CustomerTransaction.create_membership_purchase_transaction(
                                customer, customer_membership, 
                                system_remarks= "Joined Membership", 
                                transact_outlet=assigned_outlet, 
                                transact_by=assigned_by, 
                                )
    
    check_giveaway_reward_for_membership_purchase_transaction(customer, customer_transaction)
    
@customer_membership_api_bp.route('/customer/reference-code/<reference_code>/renew-membership', methods=['POST'])
@auth_token_required
#@elapsed_time_trace(trace_key="renew_membership")
def renew_membership(reference_code):
    acct_id         = request.headers.get('x-acct-id')
    outlet_key      = request.headers.get('x-outlet-key')
    logger.debug('reference_code=%s', reference_code)
    logger.debug('acct_id=%s', acct_id)
    logger.debug('outlet_key=%s', outlet_key)
    
    merchant_membership_key = request.args.get('merchant_membership_key') or request.form.get('merchant_membership_key') or request.json.get('merchant_membership_key')
    
    db_client = create_db_client(caller_info="renew_membership")
    customer_membership = None
    
    with db_client.context():
        merchant_acct = MerchantAcct.fetch(acct_id)
        customer = Customer.get_by_reference_code(reference_code, merchant_acct)
        if customer:
            merchant_membership = MerchantMembership.fetch(merchant_membership_key)
            merchant_username       = get_logged_in_api_username()
            renewed_by              = MerchantUser.get_by_username(merchant_username)
            renewed_outlet          = Outlet.fetch(outlet_key)
            #CustomerMembership.renew(customer, merchant_membership, renewed_datetime=datetime.utcnow())
            try:
                customer_membership = __renew_membership(customer, merchant_membership, renewed_by, renewed_outlet)
            except Exception as e:
                return create_rest_message(e.args[0], status_code=StatusCode.BAD_REQUEST)
            
    if customer_membership:
        return create_rest_message('Customer membership have been renew successfully', status_code=StatusCode.OK)
    else:
        return create_rest_message('Customer membership is not found', status_code=StatusCode.BAD_REQUEST)
        
    
    return create_rest_message(status_code=StatusCode.OK)    


@model_transactional(desc="renew customer membership")
def __renew_membership(customer, merchant_membership, renewed_by, renewed_outlet):
    customer_membership = CustomerMembership.renew(customer, merchant_membership, renewed_datetime=datetime.utcnow(), renewed_outlet=renewed_outlet, renewed_by=renewed_by)
    
    customer_transaction = CustomerTransaction.create_membership_purchase_transaction(
                                customer, customer_membership, 
                                system_remarks= "Renewed Membership", 
                                transact_outlet=renewed_outlet, 
                                transact_by=renewed_by, 
                                )
    
    check_giveaway_reward_for_membership_purchase_transaction(customer, customer_transaction)
    
    return  customer_membership   


'''
Created on 20 Apr 2021

@author: jacklok
'''
from trexlib.utils.google.cloud_tasks_util import create_task
from trexadmin import conf
from trexlib.conf import CHECK_CUSTOMER_ENTITLE_REWARD_TASK_URL, SYSTEM_TASK_SERVICE_CREDENTIAL_PATH, SYSTEM_TASK_GCLOUD_PROJECT_ID, SYSTEM_TASK_GCLOUD_LOCATION, SYSTEM_TASK_SERVICE_ACCOUNT_EMAIL  
import logging
from trexadmin.libs.app.reward_program.reward_program_factory import RewardProgramFactory
from trexadmin.controllers.system.system_routes import get_currency_config
from trexmodel.models.datastore.reward_models import CustomerPointReward,\
    CustomerStampReward, CustomerEntitledVoucher, VoucherRewardDetailsForUpstreamData,\
    CustomerEntitledTierRewardSummary
from datetime import datetime
from trexmodel import program_conf
from trexmodel.models.datastore.customer_model_helpers import update_customer_entiteld_voucher_summary_after_reverted_voucher,\
    update_reward_summary_with_new_reward,\
    update_prepaid_summary_with_new_prepaid,\
    update_customer_entiteld_voucher_summary_with_customer_new_voucher,\
    update_prepaid_summary_with_reverted_prepaid,\
    update_reward_summary_with_reverted_reward
from trexanalytics.bigquery_upstream_data_config import create_merchant_customer_reward_reverted_upstream_for_merchant,\
    create_merchant_customer_transaction_upstream_for_merchant,\
    create_merchant_customer_redemption_upstream_for_merchant,\
    create_merchant_customer_prepaid_upstream_for_merchant,\
    create_merchant_customer_prepaid_reverted_upstream_for_merchant,\
    create_merchant_customer_redemption_reverted_upstream_for_merchant,\
    create_merchant_sales_transaction_upstream_for_merchant
from trexmodel.models.datastore.membership_models import MerchantTierMembership
from trexmodel.models.datastore.transaction_models import CustomerTransaction,\
    SalesTransaction
from trexmodel.models.datastore.model_decorators import model_transactional
from trexmodel.models.datastore.prepaid_models import CustomerPrepaidReward
from trexadmin.libs.app.reward_program.reward_program_base import EntitledVoucherSummary
from trexmodel.models.datastore.customer_models import Customer,\
    CustomerMembership, CustomerTierMembership
from trexmodel.models.datastore.redeem_models import CustomerRedemption
from trexlib.utils.string_util import is_not_empty
from trexlib.utils.log_util import get_tracelog
from trexmodel.models.datastore.lucky_draw_models import LuckyDrawTicket
from trexmodel.models.datastore.message_model_helper import create_transaction_message

logger = logging.getLogger('debug')
#logger = logging.getLogger('debug')


def create_sales_transaction(transact_outlet=None, sales_amount=.0, tax_amount=.0, invoice_id=None, remarks=None, 
                            transact_by=None, transact_datetime=None, invoice_details=None):
    
    logger.debug('---create_sales_transaction---')
    
    logger.debug('invoice_details=%s', invoice_details)
    
    @model_transactional(desc='create_sales_transaction')
    def __start_transaction_for_customer_transaction(): 
        sales_transaction = SalesTransaction.create(
                                       transact_outlet      = transact_outlet,
                                       
                                       transact_amount      = sales_amount, 
                                       tax_amount           = tax_amount,
                                       
                                       invoice_id           = invoice_id, 
                                       remarks              = remarks,
                                       
                                       transact_by          = transact_by,
                                       
                                       transact_datetime    = transact_datetime,
                                       
                                       )
            
        #create_merchant_sales_transaction_upstream_for_merchant(sales_transaction, streamed_datetime=transact_datetime)
        
        return sales_transaction
        
    return __start_transaction_for_customer_transaction()

def create_reward_transaction(customer, transact_outlet=None, sales_amount=.0, tax_amount=.0, invoice_id=None, remarks=None, 
                            transact_by=None, transact_datetime=None, invoice_details=None):
    
    logger.debug('---create_reward_transaction---')
    
    logger.debug('invoice_details=%s', invoice_details)
    
    @model_transactional(desc='create_reward_transaction')
    def __start_transaction_for_customer_transaction():
        customer_transaction = CustomerTransaction.create_system_transaction(
                                       customer, 
                                       transact_outlet      = transact_outlet,
                                       
                                       transact_amount      = sales_amount, 
                                       tax_amount           = tax_amount,
                                       
                                       invoice_id           = invoice_id, 
                                       remarks              = remarks,
                                       
                                       transact_by          = transact_by,
                                       
                                       transact_datetime    = transact_datetime,
                                       
                                       is_sales_transaction = True,
                                       
                                       )
            
        trigger_check_reward_success = trigger_spending_reward_for_transaction(customer_transaction)
        
        logger.debug('trigger_check_reward_success=%s', trigger_check_reward_success)
        
        if trigger_check_reward_success:
            create_transaction_message(customer_transaction)
            create_merchant_customer_transaction_upstream_for_merchant(customer_transaction, streamed_datetime=transact_datetime)
        
        return customer_transaction
        
    return __start_transaction_for_customer_transaction()

def give_reward_from_sales_transaction(customer, sales_transaction):
    
    logger.debug('---give_reward_from_sales_transaction---')
    
    @model_transactional(desc='give_reward_from_sales_transaction')
    def __start_transaction(customer, sales_transaction):
        customer_transaction = CustomerTransaction.create_from_sales_transaction(customer, sales_transaction)
        
        trigger_check_reward_success = trigger_spending_reward_for_transaction(customer_transaction)
        
        logger.debug('trigger_check_reward_success=%s', trigger_check_reward_success)
        
        if trigger_check_reward_success:
            create_transaction_message(customer_transaction)
            create_merchant_customer_transaction_upstream_for_merchant(customer_transaction, streamed_datetime=sales_transaction.transact_datetime)
        
        return customer_transaction
    
    customer_transaction = CustomerTransaction.get_by_transaction_id(sales_transaction.transaction_id)
    if customer_transaction is None:
        customer_transaction = __start_transaction(customer, sales_transaction)
    
    return customer_transaction
    

def create_non_sales_system_transaction(customer, transact_datetime=None, remarks=None, system_remarks=None):
    
    logger.debug('---create_non_sales_system_transaction---')
    
    #@model_transactional(desc='create_non_sales_system_transaction')
    def __start_transaction_for_customer_transaction():
        customer_transaction = CustomerTransaction.create_system_transaction(
                                       customer, 
                                       transact_datetime    = transact_datetime,
                                       allow_to_revert      = False,
                                       is_sales_transaction = False,
                                       remarks              = remarks,
                                       system_remarks       = system_remarks,
                                       )
            
        return customer_transaction
        
    return __start_transaction_for_customer_transaction()
    
def trigger_spending_reward_for_transaction(transaction_details):
    
    if conf.CHECK_ENTITLE_REWARD_THRU_TASKQUEUE:
        transaction_key                 = transaction_details.key_in_str
        
        task_url                        = CHECK_CUSTOMER_ENTITLE_REWARD_TASK_URL
        
        logger.debug('trigger_spending_reward_for_transaction: transaction_key=%s', transaction_key)
        logger.debug('trigger_spending_reward_for_transaction: task_url=%s', task_url)
        
        queue_name      = 'giveaway-reward' 
        payload         = {
                            'transaction_key' : transaction_key
                        }
        
        logger.debug('trigger_spending_reward_for_transaction: payload=%s', payload)
        
        
                        
        create_task(task_url, queue_name, payload=payload, 
                        in_seconds      = 1, 
                        http_method     = 'POST',
                        credential_path = SYSTEM_TASK_SERVICE_CREDENTIAL_PATH, 
                        project_id      = SYSTEM_TASK_GCLOUD_PROJECT_ID,
                        location        = SYSTEM_TASK_GCLOUD_LOCATION,
                        service_email   = SYSTEM_TASK_SERVICE_ACCOUNT_EMAIL
                        )
        
        return True
    else:
        
        customer_acct       = transaction_details.transact_customer_acct
        #merchant_acct       = transaction_details.transact_merchant_acct
        check_reward_status = check_spending_reward_for_transaction(customer_acct, transaction_details)
        
        return check_reward_status

def update_customer_kpi_summary_and_transact_summary(customer_acct, transaction_details):
    
    logger.debug('---update_customer_kpi_summary_and_transact_summary---')
    
    kpi_summary = customer_acct.kpi_summary
    
    transaction_entitled_point      = .0
    transaction_entitled_prepaid    = .0
    transaction_entitled_stamp      = 0
    
    if transaction_details.entitled_reward_summary:        
        if transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_POINT):
            transaction_entitled_point = transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_POINT).get('amount') or .0
            
        if transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_STAMP):
            transaction_entitled_stamp = transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_STAMP).get('amount') or 0
    
    if transaction_details.entitled_prepaid_summary:    
        transaction_entitled_prepaid = transaction_details.entitled_prepaid_summary.get('amount') or .0    
        
    
    if kpi_summary is None:
        kpi_summary = {
                        'total_transact_amount'             : .0,
                        'total_accumulated_point'           : .0,
                        'total_accumulated_stamp'           : 0,
                        'total_accumulated_point'           : .0,
                        'total_accumulated_prepaid'         : .0,
                        }
    
    customer_acct.previous_transact_datetime    = customer_acct.last_transact_datetime
    customer_acct.last_transact_datetime        = transaction_details.transact_datetime
    
    kpi_summary['total_transact_amount']        += transaction_details.transact_amount
    kpi_summary['total_accumulated_point']      += transaction_entitled_point
    kpi_summary['total_accumulated_stamp']      += transaction_entitled_stamp
    kpi_summary['total_accumulated_prepaid']    += transaction_entitled_prepaid
    
    customer_acct.kpi_summary = kpi_summary
    customer_acct.put()
    
def update_customer_kpi_summary_from_transaction_list(customer_acct, transaction_list):
    
    logger.debug('---update_customer_kpi_summary_from_transaction_list---')
    
    kpi_summary = customer_acct.kpi_summary
    
    transaction_entitled_point      = .0
    transaction_entitled_prepaid    = .0
    transaction_entitled_stamp      = 0
    transaction_transact_amount     = .0
    
    for transaction_details in transaction_list:
        
        if transaction_details.entitled_reward_summary:        
            if transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_POINT):
                transaction_entitled_point += transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_POINT).get('amount') or .0
                
            if transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_STAMP):
                transaction_entitled_stamp += transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_STAMP).get('amount') or 0
        
        if transaction_details.entitled_prepaid_summary:    
            transaction_entitled_prepaid += transaction_details.entitled_prepaid_summary.get('amount') or .0    
            
        transaction_transact_amount+=transaction_details.transact_amount
        
        logger.debug('update_customer_kpi_summary_from_transaction_list debug: transaction_entitled_point=%s', transaction_entitled_point)
        logger.debug('update_customer_kpi_summary_from_transaction_list debug: transaction_entitled_prepaid=%s', transaction_entitled_prepaid)
        logger.debug('update_customer_kpi_summary_from_transaction_list debug: transaction_entitled_stamp=%s', transaction_entitled_stamp)
        logger.debug('update_customer_kpi_summary_from_transaction_list debug: transaction_transact_amount=%s', transaction_transact_amount)
        
        if kpi_summary is None:
            kpi_summary = {
                            'total_transact_amount'             : .0,
                            'total_accumulated_point'           : .0,
                            'total_accumulated_stamp'           : 0,
                            'total_accumulated_point'           : .0,
                            'total_accumulated_prepaid'         : .0,
                            }
        if transaction_details.transact_datetime > customer_acct.last_transact_datetime:    
            customer_acct.previous_transact_datetime    = customer_acct.last_transact_datetime
            customer_acct.last_transact_datetime        = transaction_details.transact_datetime
    
    kpi_summary['total_transact_amount']        += transaction_transact_amount
    kpi_summary['total_accumulated_point']      += transaction_entitled_point
    kpi_summary['total_accumulated_stamp']      += transaction_entitled_stamp
    kpi_summary['total_accumulated_prepaid']    += transaction_entitled_prepaid
    
    logger.debug('update_customer_kpi_summary_from_transaction_list debug: kpi_summary=%s', kpi_summary)
    
    customer_acct.kpi_summary = kpi_summary
    customer_acct.put()    
    
    logger.debug('update_customer_kpi_summary_from_transaction_list debug: customer_acct.kpi_summary after updated=%s', customer_acct.kpi_summary)
    

def check_for_tier_membership_upgrade_downgrade(customer_acct, merchant_acct, transaction_details):
    
    logger.debug('---check_for_tier_membership_upgrade_downgrade---')
    if merchant_acct.is_tier_membership_configured:
        merchant_tier_membership_list   = MerchantTierMembership.list_by_merchant_acct(merchant_acct)
        existing_tier_membership_key    = customer_acct.tier_membership_key
        existing_tier_membership        = None
        membership_to_assign            = None
        
        customer_kpi_summary            = customer_acct.kpi_summary
        
        total_transact_amount           = customer_kpi_summary['total_transact_amount']
        total_accumulated_point         = customer_kpi_summary['total_accumulated_point']
        total_accumulated_stamp         = customer_kpi_summary['total_accumulated_stamp'] 
        total_accumulated_prepaid       = customer_kpi_summary['total_accumulated_prepaid']
        
        logger.debug('check_for_tier_membership_upgrade_downgrade: total_transact_amount=%s', total_transact_amount)
        logger.debug('check_for_tier_membership_upgrade_downgrade: total_accumulated_point=%s', total_accumulated_point)
        logger.debug('check_for_tier_membership_upgrade_downgrade: total_accumulated_stamp=%s', total_accumulated_stamp)
        logger.debug('check_for_tier_membership_upgrade_downgrade: total_accumulated_prepaid=%s', total_accumulated_prepaid)
        
        existing_membership_qualification_details                   = {}
        highest_entitle_qualification_details                       = {}
        no_membership_or_existing_membership_is_auto_assigned       = False
        
        if existing_tier_membership_key:
            existing_tier_membership        = MerchantTierMembership.fetch(existing_tier_membership_key)
            
            if existing_tier_membership:
                if existing_tier_membership.archived==True:
                    existing_tier_membership = None
                    no_membership_or_existing_membership_is_auto_assigned = True
                else:
                    if existing_tier_membership.entitle_qualification_type == program_conf.MEMBERSHIP_ENTITLE_QUALIFICATION_TYPE_AUTO_ASSIGN:
                        no_membership_or_existing_membership_is_auto_assigned = True
                    else:
                        existing_membership_qualification_details[existing_tier_membership.entitle_qualification_type] = existing_tier_membership.entitle_qualification_value
        else:
            no_membership_or_existing_membership_is_auto_assigned = True
                
        for membership in merchant_tier_membership_list:
            logger.debug('checking tier membership %s, entitle_qualification_type=%s', membership.label, membership.entitle_qualification_type)
            
            if membership.entitle_qualification_type == program_conf.MEMBERSHIP_ENTITLE_QUALIFICATION_TYPE_AUTO_ASSIGN:
                if membership_to_assign is None:
                    membership_to_assign = membership
                    
                    logger.debug('Found auto assign tier membership')
                    
            if membership.entitle_qualification_type == program_conf.MEMBERSHIP_ENTITLE_QUALIFICATION_TYPE_EXCEED_SPENDING_AMOUNT:
                logger.debug('checking tier membership is based on spending amount')
                if transaction_details.transact_amount >= membership.entitle_qualification_value:
                    highest_entitle_qualification_value =  highest_entitle_qualification_details.get(membership.entitle_qualification_type) or 0
                    if membership.entitle_qualification_value > highest_entitle_qualification_value:
                        membership_to_assign = membership
                        highest_entitle_qualification_details[membership.entitle_qualification_type] = membership.entitle_qualification_value
                        
                        logger.debug('Found %s tier membership to assign', membership.label)
                
                    
            if membership.entitle_qualification_type == program_conf.MEMBERSHIP_ENTITLE_QUALIFICATION_TYPE_ACCUMULATED_SPENDING_AMOUNT:
                logger.debug('checking tier membership is based on accumulated spending amount, where amount is %s', membership.entitle_qualification_value)
                if total_transact_amount >= membership.entitle_qualification_value:
                    highest_entitle_qualification_value =  highest_entitle_qualification_details.get(membership.entitle_qualification_type) or 0
                    if membership.entitle_qualification_value > highest_entitle_qualification_value:
                        membership_to_assign = membership
                        highest_entitle_qualification_details[membership.entitle_qualification_type] = membership.entitle_qualification_value
                        
                        logger.debug('Found %s tier membership to assign', membership.label)
                        
                    
            if membership.entitle_qualification_type == program_conf.MEMBERSHIP_ENTITLE_QUALIFICATION_TYPE_ACCUMULATED_POINT_AMOUNT:
                logger.debug('checking tier membership is based on accumulated point amount, where amount is %s', membership.entitle_qualification_value)
                if total_accumulated_point >= membership.entitle_qualification_value:
                    highest_entitle_qualification_value =  highest_entitle_qualification_details.get(membership.entitle_qualification_type) or 0
                    if membership.entitle_qualification_value > highest_entitle_qualification_value:
                        membership_to_assign = membership
                        highest_entitle_qualification_details[membership.entitle_qualification_type] = membership.entitle_qualification_value
                        
                        logger.debug('Found %s tier membership to assign', membership.label)
                
                                        
            if membership.entitle_qualification_type == program_conf.MEMBERSHIP_ENTITLE_QUALIFICATION_TYPE_ACCUMULATED_STAMP_AMOUNT:
                logger.debug('checking tier membership is based on accumulated stamp amount, where amount is %s', membership.entitle_qualification_value)
                if total_accumulated_stamp >= membership.entitle_qualification_value:
                    highest_entitle_qualification_value =  highest_entitle_qualification_details.get(membership.entitle_qualification_type) or 0
                    if membership.entitle_qualification_value > highest_entitle_qualification_value:
                        membership_to_assign = membership
                        highest_entitle_qualification_details[membership.entitle_qualification_type] = membership.entitle_qualification_value
                        
                        logger.debug('Found %s tier membership to assign', membership.label)
                        
                    
            if membership.entitle_qualification_type == program_conf.MEMBERSHIP_ENTITLE_QUALIFICATION_TYPE_ACCUMULATED_PREPAID_AMOUNT:
                logger.debug('checking tier membership is based on accumulated prepaid amount, where amount is %s', membership.entitle_qualification_value)
                if total_accumulated_prepaid >= membership.entitle_qualification_value:
                    highest_entitle_qualification_value =  highest_entitle_qualification_details.get(membership.entitle_qualification_type) or 0
                    if membership.entitle_qualification_value > highest_entitle_qualification_value:
                        membership_to_assign = membership
                        highest_entitle_qualification_details[membership.entitle_qualification_type] = membership.entitle_qualification_value
                        
                        logger.debug('Found %s tier membership to assign', membership.label)
                        
        
        if membership_to_assign:
            logger.debug('membership_to_assign=%s, ', membership_to_assign.label)
            
            
            is_membership_change = membership_to_assign.key_in_str!=existing_tier_membership_key
            
            logger.debug('is_membership_change=%s', is_membership_change)
            
            if existing_tier_membership:
                logger.debug('existing_tier_membership=%s', existing_tier_membership.label)
                
                if is_membership_change:
                    logger.debug('check_for_tier_membership_upgrade_downgrade: found upgrading tier membership to assign')
                    
                    CustomerTierMembership.change(customer_acct,
                                              membership_to_assign, 
                                              transaction_details,
                                              entitled_datetime=transaction_details.transact_datetime,
                                              )
                    
                    
                    
                    logger.debug('existing_membership_qualification_details=%s', existing_membership_qualification_details)
                    logger.debug('no_membership_or_existing_membership_is_auto_assigned=%s', no_membership_or_existing_membership_is_auto_assigned)
                    
                    check_reward_for_new_membership(customer_acct, transaction_details, membership_to_assign)
                    
                    '''
                    #if existing_membership_qualification_details.get(membership_to_assign.entitle_qualification_type):
                    if no_membership_or_existing_membership_is_auto_assigned:
                        logger.debug('going to check reward for new tier membership')
                        check_reward_for_new_membership(customer_acct, transaction_details, membership_to_assign)
                    else:
                        if highest_entitle_qualification_details.get(membership_to_assign.entitle_qualification_type) > existing_membership_qualification_details.get(membership_to_assign.entitle_qualification_type):
                            logger.debug('Membership is upgrading')
                            check_reward_for_new_membership(customer_acct, transaction_details, membership_to_assign)
                        else:
                            logger.debug('Membership is downgrading')
                    '''
                    
                    
                else:
                    logger.debug('check_for_tier_membership_upgrade_downgrade: remain as existing tier membership')
            else:
                logger.debug('check_for_tier_membership_upgrade_downgrade: found new tier membership to assign')
                CustomerTierMembership.create(customer_acct, 
                                              membership_to_assign, 
                                              transaction_details,
                                              entitled_datetime=transaction_details.transact_datetime,
                                              
                                              )
                #customer_acct.tier_membership = new_tier_membership_key
                
                check_reward_for_new_membership(customer_acct, transaction_details, membership_to_assign)
                
                #customer_acct.put()
        else:
            if existing_tier_membership:
                #should be downgrade and remove membership
                customer_acct.tier_membership           = None
                customer_acct.put()
                logger.debug('check_for_tier_membership_upgrade_downgrade: not tier membership to upgrade, thus remain as existing tier membership')
            else:
                logger.debug('check_for_tier_membership_upgrade_downgrade: not yet entitle any tier membership')
    else:
        logger.debug('no tier membership is configured thus ignore')

def check_reward_for_new_membership(customer_acct, transaction_details, new_membership):
    
    logger.debug('---check_reward_for_new_membership---')
    
    merchant_acct                           = transaction_details.transact_merchant_acct
    merchant_program_configuration_list     = merchant_acct.published_program_configuration.get('programs')
    
    membership_program_configuration_list = []
    
    logger.debug('Found merchant_program_configuration_list count=%d', len(merchant_program_configuration_list))
    
    for program_configuration in merchant_program_configuration_list:
        
        program_desc = program_configuration.get('desc')
        program_reward_base = program_configuration.get('reward_base')
        
        logger.debug('check_reward_for_new_membership: program_desc=%s', program_desc)
        logger.debug('check_reward_for_new_membership: program_reward_base=%s', program_reward_base)
        
        if  program_reward_base== program_conf.REWARD_BASE_ON_GIVEAWAY:
            
            program_giveaway_method = program_configuration.get('giveaway_method')
            
            logger.debug('check_reward_for_new_membership: program_giveaway_method=%s', program_giveaway_method)
            
            if  program_giveaway_method== program_conf.PROGRAM_REWARD_GIVEAWAY_METHOD_SYSTEM:
                logger.debug('found giveaway method is system')
                giveaaway_system_settings = program_configuration.get('program_settings').get('giveaway_system_settings')
                
                logger.debug('check_reward_for_new_membership: giveaaway_system_settings=%s', giveaaway_system_settings)
                
                if giveaaway_system_settings is not None and giveaaway_system_settings.get('giveaway_system_condition') == program_conf.GIVEAWAY_SYSTEM_CONDITION_NEW_MEMBERSHIP:
                    logger.debug('giveaway tier memberships=%s', giveaaway_system_settings.get('giveaway_tier_memberships'))
                    if giveaaway_system_settings.get('giveaway_tier_memberships'):
                        if new_membership.key_in_str in giveaaway_system_settings.get('giveaway_tier_memberships'):
                            membership_program_configuration_list.append(program_configuration)
    
    logger.debug('membership_program_configuration_list=%s', membership_program_configuration_list)
    
    if membership_program_configuration_list:
        currency_code   = merchant_acct.currency_code
        currency_config = get_currency_config(currency_code)
        return RewardProgramFactory().get_giveaway_reward(customer_acct, transaction_details, 
                                                                            program_configuration_list  = membership_program_configuration_list, 
                                                                            currency                    = currency_config
                                                                            )
    
    
def check_tier_reward_for_transaction(customer_acct, transaction_details, program_configuration_list=None):
    
    logger.debug('---check_tier_reward_for_transaction---')
    
    merchant_acct   = transaction_details.transact_merchant_acct
    currency_code   = merchant_acct.currency_code
    
    currency_config = get_currency_config(currency_code)
    
    return RewardProgramFactory().get_tier_reward(customer_acct, 
                                                            transaction_details, 
                                                            program_configuration_list=program_configuration_list, 
                                                            currency=currency_config)    

    
def check_spending_reward_for_transaction(customer_acct, transaction_details, program_configuration_list=None):
    merchant_acct   = transaction_details.transact_merchant_acct
    currency_code   = merchant_acct.currency_code
    
    currency_config = get_currency_config(currency_code)
    
    give_reward_status = RewardProgramFactory().get_spending_reward(customer_acct, 
                                                            transaction_details, 
                                                            program_configuration_list=program_configuration_list, 
                                                            currency=currency_config)
    
    
    if give_reward_status:
        #check if the giveaway reward format match tier reward program setting
        check_tier_reward_for_transaction(customer_acct, transaction_details)
        update_customer_kpi_summary_and_transact_summary(customer_acct, transaction_details)
        check_for_tier_membership_upgrade_downgrade(customer_acct, merchant_acct, transaction_details)
        check_entitled_lucky_draw_ticket_for_transaction(customer_acct, merchant_acct, transaction_details)
        
    return give_reward_status

def check_entitled_lucky_draw_ticket_for_transaction(customer_acct, merchant_acct, transaction_details):
    
    if merchant_acct.lucky_draw_program_count>0:
        entries_list = LuckyDrawTicket.create_for_customer_from_sales_amount(customer_acct, 
                                                                             sales_amount=transaction_details.transact_amount, 
                                                                             merchant_acct=merchant_acct, 
                                                                             transact_outlet=transaction_details.transact_outlet_entity)
        ticket_configurations_list = []
        logger.debug('entries_list count=%d', len(entries_list))
        
        for e in entries_list:
            ticket_configurations_list.append(e.to_configuration())
        
        if ticket_configurations_list:
            Customer.add_new_tickets_list_into_lucky_draw_ticket_summary(customer_acct, ticket_configurations_list)
            transaction_details.update_entitled_lucky_draw_ticket_summary(ticket_configurations_list)
            
            
def check_giveaway_reward_for_transaction(customer_acct, transaction_details, program_configuration_list=None, reward_set=1):
    merchant_acct                   = transaction_details.transact_merchant_acct
    program_configuration_list      = merchant_acct.program_configuration_list if program_configuration_list is None else program_configuration_list
    currency_code                   = merchant_acct.currency_code
    
    currency_config                 = get_currency_config(currency_code)
    
    give_reward_status =  RewardProgramFactory().get_giveaway_reward(customer_acct, 
                                                            transaction_details, 
                                                            program_configuration_list=program_configuration_list, 
                                                            currency=currency_config, 
                                                            reward_set=reward_set)
    
    if give_reward_status:
        #check if the giveaway reward format match tier reward program setting
        check_tier_reward_for_transaction(customer_acct, transaction_details)
        update_customer_kpi_summary_and_transact_summary(customer_acct, transaction_details)
        check_for_tier_membership_upgrade_downgrade(customer_acct, merchant_acct, transaction_details)
    
    return give_reward_status

def check_giveaway_reward_for_membership_purchase_transaction(customer_acct, transaction_details, program_configuration_list=None, reward_set=1):
    logger.debug('---check_giveaway_reward_for_membership_purchase_transaction---')
    merchant_acct                   = transaction_details.transact_merchant_acct
    program_configuration_list      = merchant_acct.program_configuration_list if program_configuration_list is None else program_configuration_list
    
    currency_code   = merchant_acct.currency_code
    
    currency_config = get_currency_config(currency_code)
    
    give_reward_status =  RewardProgramFactory().get_purchase_membership_giveaway_reward(customer_acct, 
                                                            transaction_details, 
                                                            program_configuration_list=program_configuration_list, 
                                                            currency=currency_config, 
                                                            reward_set=reward_set)
    
    '''
    if give_reward_status:
        #check if the giveaway reward format match tier reward program setting
        check_tier_reward_for_transaction(customer_acct, transaction_details)
    '''
    return give_reward_status

def redeem_reward_transaction(customer, redeem_outlet=None, reward_format=None, reward_amount=.0, invoice_id=None, remarks=None, 
                            redeemed_by=None, redeemed_datetime=None, transaction_id=None,
                            redeemed_voucher_keys_list=None
                            ):
    logger.debug('---redeem_reward_transaction---')
    
    if reward_format in program_conf.REWARD_FORMAT_SET:
    
        @model_transactional(desc='redeem_reward_transaction')
        def __start_transaction_for_customer_transaction():
            
            logger.debug('reward_format=%s', reward_format)
            
            customer_redemption = CustomerRedemption.create(customer, reward_format, redeem_outlet, 
                                      redeemed_amount               = reward_amount,
                                      redeemed_voucher_keys_list    = redeemed_voucher_keys_list,                      
                                      invoice_id                    = invoice_id, 
                                      transaction_id                = transaction_id,
                                      remarks                       = remarks, 
                                      redeemed_by                   = redeemed_by, 
                                      redeemed_datetime             = redeemed_datetime)
            
            if customer_redemption:
                create_merchant_customer_redemption_upstream_for_merchant(customer_redemption, streamed_datetime=redeemed_datetime)
            
            return customer_redemption
            
        return __start_transaction_for_customer_transaction()
    else:
        raise Exception("Invalid reward format")

def update_customer_reward_summary(customer):
    entitled_point_details_list     = CustomerPointReward.list_by_customer(customer)
    entitled_stamp_details_list     = CustomerStampReward.list_by_customer(customer)      
    
    entitled_reward_summary         = {}
    
    logger.debug('update_customer_reward_summary: entitled_point_details_list=%s', entitled_point_details_list)
    logger.debug('update_customer_reward_summary: entitled_stamp_details_list=%s', entitled_stamp_details_list)
    
    for reward in  entitled_point_details_list:
        entiteld_reward_details = reward.to_reward_summary()
        entitled_reward_summary = update_reward_summary_with_new_reward(entitled_reward_summary, entiteld_reward_details)
        
    for reward in  entitled_stamp_details_list:
        entiteld_reward_details = reward.to_reward_summary()
        entitled_reward_summary = update_reward_summary_with_new_reward(entitled_reward_summary, entiteld_reward_details)
     
    logger.debug('entitled_reward_summary=%s', entitled_reward_summary) 
        
    customer.reward_summary    = entitled_reward_summary
    
    customer.put()
    
def update_customer_prepaid_summary(customer):
    prepaid_reward_list     = CustomerPrepaidReward.list_by_customer(customer)    
    prepaid_summary         = {}
    
    for prepaid in  prepaid_reward_list:
        prepaid = prepaid.to_prepaid_summary()
        prepaid_summary = update_prepaid_summary_with_new_prepaid(prepaid_summary, prepaid)
    
    customer.prepaid_summary = prepaid_summary
    customer.put()
    
def update_customer_memberships_list(customer):
    customer_memberships_list     = CustomerMembership.list_all_by_customer(customer)
    memberships_list              = []
    
    
    for customer_membership in  customer_memberships_list:
        if customer_membership.is_valid():
            memberships_list.append(customer_membership.merchant_membership_key)
        
        
    customer.memberships_list    = memberships_list
    
    customer.put()    

def revert_transaction(transaction_details, reverted_by, reverted_datetime=None, create_upstream=True):
    merchant_acct                       = transaction_details.transact_merchant_acct
    transaction_id                      = transaction_details.transaction_id
    customer_acct                       = transaction_details.transact_customer_acct
        
    if transaction_details.is_membership_purchase:
        #remove membership
        
        merchant_membership_key         = transaction_details.purchased_merchant_membership_key
        merchant_membership             = transaction_details.purchased_merchant_membership_entity
        purchased_customer_membership   = CustomerMembership.get_by_customer_and_merchant_membership(customer_acct, merchant_membership)
        if purchased_customer_membership:
            try:
                 
                purchased_customer_membership.delete()
            except:
                logger.error('Failed due to %s', get_tracelog())
        
        else:
            logger.warn('Customer membership is not found')
        
        if is_not_empty(customer_acct.memberships_list):
            customer_acct.memberships_list.remove(merchant_membership_key)
        
        customer_acct.put()
        
        return True
        
    elif transaction_details.is_membership_renew:
        #revert renewal
        renewed_customer_membership = transaction_details.purchased_customer_membership_entity
        renewed_customer_membership.revert_renewal()    
        customer_acct.put()
        
        return True
    else:    
    
        is_stamp_entitled_in_transaction    = transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_STAMP) is not None
        is_point_entitled_in_transaction    = transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_POINT) is not None
        is_voucher_entitled_in_transaction  = transaction_details.entitled_voucher_summary is not None
        is_prepaid_entitled_in_transaction  = transaction_details.entitled_prepaid_summary is not None
        
        
        logger.debug('transaction_details.entitled_reward_summary=%s', transaction_details.entitled_reward_summary)
        logger.debug('transaction_details.entitled_prepaid_summary=%s', transaction_details.entitled_prepaid_summary)
        
        
        reverting_point_details_list        = CustomerPointReward.list_by_transaction_id(transaction_id) if is_point_entitled_in_transaction else []
        reverting_stamp_details_list        = CustomerStampReward.list_by_transaction_id(transaction_id) if is_stamp_entitled_in_transaction else []
        reverting_vouchers_list             = CustomerEntitledVoucher.list_by_transaction_id(transaction_id) if is_voucher_entitled_in_transaction else []
        reverting_prepaid_list              = CustomerPrepaidReward.list_by_transaction_id(transaction_id) if is_prepaid_entitled_in_transaction else []
        
        is_transaction_reward_used          = False
        customer_reward_summary             = customer_acct.reward_summary or {}
        customer_prepaid_summary            = customer_acct.prepaid_summary or {}
        entitled_voucher_summary            = customer_acct.entitled_voucher_summary or {}
        
        
        if reverted_datetime is None:
            reverted_datetime = datetime.utcnow()
        
        logger.debug('revert_transaction: transaction_id=%s', transaction_id)
        logger.debug('revert_transaction: reverted_datetime=%s', reverted_datetime)
        logger.debug('revert_transaction: reverting_point_details_list count=%s', len(reverting_point_details_list))
        logger.debug('revert_transaction: reverting_stamp_details_list count=%s', len(reverting_stamp_details_list))
        logger.debug('revert_transaction: reverting_vouchers_list count=%s', len(reverting_vouchers_list))
        logger.debug('revert_transaction: reverting_prepaid_list count=%s', len(reverting_prepaid_list))
        
        for p in reverting_point_details_list:
            if p.is_used:
                is_transaction_reward_used = True
                break
        
        if is_transaction_reward_used is False:
            for p in reverting_stamp_details_list:
                if p.is_used:
                    is_transaction_reward_used = True
                    break
                
        if is_transaction_reward_used is False:
            for p in reverting_vouchers_list:
                if p.is_used:
                    is_transaction_reward_used = True
                    break
                
        if is_transaction_reward_used is False:
            for p in reverting_prepaid_list:
                if p.is_used:
                    is_transaction_reward_used = True
                    break        
        
        logger.debug('revert: is_transaction_reward_used=%s', is_transaction_reward_used)
        
        if is_transaction_reward_used is False:
            
            logger.debug('Going to revert transaction')
            
            reverted_by_key                                = reverted_by.create_ndb_key()
            transaction_details.is_revert                  = True
            transaction_details.reverted_datetime          = reverted_datetime
            transaction_details.reverted_by                = reverted_by_key
            transaction_details.reverted_by_username       = reverted_by.username
            transaction_details.put()
            
            for p in reverting_point_details_list:
                if p.is_valid:
                    p.status                        = program_conf.REWARD_STATUS_REVERTED
                    p.reverted_datetime             = reverted_datetime
                    p.reverted_by                   = reverted_by_key
                    p.reverted_by_username          = reverted_by.username
                    p.put()
                    
                    customer_reward_summary = update_reward_summary_with_reverted_reward(customer_reward_summary, p.to_reward_summary())
                    if create_upstream:
                        create_merchant_customer_reward_reverted_upstream_for_merchant(transaction_details, p, reverted_datetime)
                    
                
            for p in reverting_stamp_details_list:
                if p.is_valid:
                    p.status                        = program_conf.REWARD_STATUS_REVERTED
                    p.reverted_datetime             = reverted_datetime
                    p.reverted_by                   = reverted_by_key
                    p.reverted_by_username          = reverted_by.username
                    p.put()
                    
                    customer_reward_summary = update_reward_summary_with_reverted_reward(customer_reward_summary, p.to_reward_summary())
                    if create_upstream:
                        create_merchant_customer_reward_reverted_upstream_for_merchant(transaction_details, p, reverted_datetime)
                
            for p in reverting_vouchers_list:
                if p.is_valid:
                    p.status                        = program_conf.REWARD_STATUS_REVERTED
                    p.reverted_datetime             = reverted_datetime
                    p.reverted_by                   = reverted_by_key
                    p.reverted_by_username          = reverted_by.username
                    p.put()
                    
                    entitled_voucher_summary = update_customer_entiteld_voucher_summary_after_reverted_voucher(entitled_voucher_summary, p)
                    
                    voucher_key         = p.entitled_voucher_key
                    expiry_date         = p.expiry_date
                    rewarded_datetime   = p.rewarded_datetime
                    
                    voucher_reward_brief = VoucherRewardDetailsForUpstreamData(voucher_key, 1, expiry_date, rewarded_datetime)
                    if create_upstream:
                        create_merchant_customer_reward_reverted_upstream_for_merchant(transaction_details, voucher_reward_brief, reverted_datetime)
            
            for p in reverting_prepaid_list:
                if p.is_valid:
                    p.status                        = program_conf.REWARD_STATUS_REVERTED
                    p.reverted_datetime             = reverted_datetime
                    p.reverted_by                   = reverted_by_key
                    p.reverted_by_username          = reverted_by.username
                    p.put()
                    
                    customer_prepaid_summary = update_prepaid_summary_with_reverted_prepaid(customer_prepaid_summary, p.to_prepaid_summary())
                    if create_upstream:
                        #create_merchant_customer_prepaid_reverted_upstream_for_merchant(transaction_details, p, reverted_datetime)
                        create_merchant_customer_reward_reverted_upstream_for_merchant(transaction_details, p, reverted_datetime)
            
            
            
            if is_stamp_entitled_in_transaction or is_point_entitled_in_transaction:
                #update_customer_reward_summary(customer_acct)
                customer_acct.reward_summary           = customer_reward_summary
            
            if is_prepaid_entitled_in_transaction:
                #update_customer_prepaid_summary(customer_acct)
                customer_acct.prepaid_summary           = customer_prepaid_summary
            
            if is_voucher_entitled_in_transaction:
                customer_acct.entitled_voucher_summary  = entitled_voucher_summary
            
            total_transact_amount       = .0
            total_accumulated_point     = .0
            total_accumulated_stamp     = 0
            total_accumulated_topup     = .0
            total_accumulated_prepaid   = .0
            
            
            if customer_acct.kpi_summary:
                total_transact_amount           = customer_acct.kpi_summary.get('total_transact_amount') or .0
                total_accumulated_point         = customer_acct.kpi_summary.get('total_accumulated_point') or .0
                total_accumulated_stamp         = customer_acct.kpi_summary.get('total_accumulated_stamp') or 0
                total_accumulated_topup         = customer_acct.kpi_summary.get('total_accumulated_topup') or .0
                total_accumulated_prepaid       = customer_acct.kpi_summary.get('total_accumulated_prepaid') or .0
            
            else:
                customer_acct.kpi_summary = {}
            
            transaction_entitled_point      = .0
            transaction_entitled_stamp      = 0
            transaction_topup_prepaid       = .0
            transaction_entitled_prepaid    = .0
            
            logger.debug('is_point_entitled_in_transaction=%s', is_point_entitled_in_transaction)
            logger.debug('is_stamp_entitled_in_transaction=%s', is_stamp_entitled_in_transaction)
            logger.debug('is_prepaid_entitled_in_transaction=%s', is_prepaid_entitled_in_transaction)
            
            if is_point_entitled_in_transaction:
                logger.debug('transaction_details entitled point summary=%s', transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_POINT))
                transaction_entitled_point = transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_POINT).get('amount') or 0
                
            if is_stamp_entitled_in_transaction:
                logger.debug('transaction_details entitled stamp summary=%s', transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_STAMP))
                transaction_entitled_stamp = transaction_details.entitled_reward_summary.get(program_conf.REWARD_FORMAT_STAMP).get('amount') or 0
                
            if is_prepaid_entitled_in_transaction:
                logger.debug('transaction_details entitled prepaid summary=%s', transaction_details.entitled_prepaid_summary)
                transaction_topup_prepaid = transaction_details.entitled_prepaid_summary.get('amount') or .0
                
            customer_acct.last_transact_datetime    = customer_acct.previous_transact_datetime
            total_transact_amount                   -= transaction_details.transact_amount
            total_accumulated_point                 -= transaction_entitled_point
            total_accumulated_stamp                 -= transaction_entitled_stamp
            total_accumulated_topup                 -= transaction_topup_prepaid
            total_accumulated_prepaid               -= transaction_entitled_prepaid
            
            if total_transact_amount<0:
                total_transact_amount = .0
                
            if total_accumulated_point<0:
                total_accumulated_point = .0
                
            if total_accumulated_stamp<0:
                total_accumulated_stamp = 0        
                
            if total_accumulated_topup<0:
                total_accumulated_topup = .0
                
            if total_accumulated_prepaid<0:
                total_accumulated_prepaid = .0        
                
            customer_acct.kpi_summary['total_transact_amount']          = total_transact_amount
            customer_acct.kpi_summary['total_accumulated_point']        = total_accumulated_point
            customer_acct.kpi_summary['total_accumulated_stamp']        = total_accumulated_stamp
            customer_acct.kpi_summary['total_accumulated_topup']        = total_accumulated_topup
            customer_acct.kpi_summary['total_accumulated_prepaid']      = total_accumulated_prepaid
            
            
            customer_acct.put()
             
            check_for_tier_membership_upgrade_downgrade(customer_acct, merchant_acct, transaction_details) 
            
            __revert_tier_reward(customer_acct, transaction_details)
            
            
            return True
        
        else:
            raise Exception('Transaction reward have been used')
    

def __revert_tier_reward(customer, transaction_details):
    customer_tier_reward_summary_list    = CustomerEntitledTierRewardSummary.list_tier_reward_summary_by_customer(customer)
    transaction_id = transaction_details.transaction_id
    
    logger.debug('__revert_tier_reward debug: transation id to revert tier reward = %s', transaction_id)
    
    if is_not_empty(customer_tier_reward_summary_list):
        for customer_tier_reward_summary in customer_tier_reward_summary_list:
            tier_summary = customer_tier_reward_summary.tier_summary
            is_tier_summary_reverted = False
            logger.debug('tier_summary before =%s', tier_summary)
            
            for tier in tier_summary.get('tiers'):
                unlock_value = tier.get('unlock_value')
                
                #for unlock_source in tier.get('unlock_source_details'):
                unlock_sources_list = []
                for index, unlock_source in enumerate(tier.get('unlock_source_details')):    
                    if unlock_source.get('transaction_id') == transaction_id:
                        logger.debug('__revert_tier_reward debug: Found unlock source based on transaction id, tier no=%d, unlock amount=%d', index, unlock_source.get('amount'))
                        unlock_value -= unlock_source.get('amount')
                        
                        tier['unlock_status']   = False
                        tier['unlock_value']    = unlock_value
                        
                        is_tier_summary_reverted = True
                    else:
                        unlock_sources_list.append(unlock_source)
                
                tier['unlock_source_details'] = unlock_sources_list    
                
            logger.debug('tier_summary after =%s', tier_summary)
            
            if is_tier_summary_reverted:
                customer_tier_reward_summary.put() 
    else:
        logger.debug('__revert_tier_reward debug: not found any tier reward cycle for customer')
        
        
def create_topup_prepaid_transaction(customer, prepaid_program, topup_outlet=None, topup_amount=.0, invoice_id=None, remarks=None, system_remarks=None,
                            topup_by=None):
    
    logger.debug('---create_topup_prepaid_transaction---')
    
    transact_datetime   = datetime.now()
    prepaid_summary     = {}
    
    @model_transactional(desc='create_topup_prepaid_transaction')
    def __start_transaction_for_customer_transaction():
        
        logger.debug('topup_amount=%s', topup_amount)
        
        customer_transaction = CustomerTransaction.create_system_transaction(
                                       customer, 
                                       transact_outlet      = topup_outlet,
                                       
                                       transact_amount      = topup_amount, 
                                       
                                       invoice_id           = invoice_id, 
                                       remarks              = remarks,
                                       system_remarks       = system_remarks,
                                       
                                       transact_by          = topup_by,
                                       
                                       is_sales_transaction = True,
                                       
                                       )
            
        prepaid_topup_reward = CustomerPrepaidReward.topup(customer, topup_amount, prepaid_program, topup_outlet=topup_outlet, 
                                                           topup_by=topup_by, transaction_id=customer_transaction.transaction_id)
        
        
        if prepaid_topup_reward:
            customer_transaction.entitled_prepaid_summary = prepaid_topup_reward.to_prepaid_summary()
            customer_transaction.put()
            
            prepaid_summary = customer.prepaid_summary
            
            logger.debug('Customer existing , prepaid_summary=%s', prepaid_summary)
            
            prepaid_summary = update_prepaid_summary_with_new_prepaid(prepaid_summary, prepaid_topup_reward.to_prepaid_summary())
            
            logger.debug('After new topup, customer(%s) prepaid_summary=%s', customer.name,prepaid_summary)
            
            customer.prepaid_summary = prepaid_summary
            customer.put()
            
            update_customer_kpi_summary_and_transact_summary(customer, customer_transaction)
            
            create_transaction_message(customer_transaction)
            create_merchant_customer_prepaid_upstream_for_merchant(customer_transaction, prepaid_topup_reward, streamed_datetime=transact_datetime)
            create_merchant_customer_transaction_upstream_for_merchant(customer_transaction, streamed_datetime=transact_datetime)
        
        return (customer_transaction, prepaid_summary)
        
    return __start_transaction_for_customer_transaction()  


def update_customer_all_entitled_reward_summary(customer):
    point_reward_list       = CustomerPointReward.list_by_customer(customer)
    stamp_reward_list       = CustomerStampReward.list_by_customer(customer)
    vouchers_list           = CustomerEntitledVoucher.list_by_customer(customer)
    prepaid_reward_list     = CustomerPrepaidReward.list_by_customer(customer) 
    
    customer_reward_summary              = {}
    customer_prepaid_summary             = {}
    customer_entitled_voucher_summary    = {}
    
    if point_reward_list:
        for point in point_reward_list:
            customer_reward_summary = update_reward_summary_with_new_reward(customer_reward_summary, point.to_reward_summary())
            
    if stamp_reward_list:
        for stamp in stamp_reward_list:
            customer_reward_summary = update_reward_summary_with_new_reward(customer_reward_summary, stamp.to_reward_summary())
            
    if vouchers_list:
        for voucher in vouchers_list:
            customer_entitled_voucher_summary = update_customer_entiteld_voucher_summary_with_customer_new_voucher(customer_entitled_voucher_summary, voucher)
                    
    
    if prepaid_reward_list:    
        for prepaid in prepaid_reward_list:
            customer_prepaid_summary = update_prepaid_summary_with_new_prepaid(customer_prepaid_summary, prepaid.to_prepaid_summary())
            
    customer.reward_summary             = customer_reward_summary
    customer.prepaid_summary            = customer_prepaid_summary
    customer.entitled_voucher_summary   = customer_entitled_voucher_summary
    
    customer.put()
    
def update_customer_entitled_reward_summary(customer):
    point_reward_list       = CustomerPointReward.list_by_customer(customer)
    stamp_reward_list       = CustomerStampReward.list_by_customer(customer)
    
    customer_reward_summary              = {}
    if point_reward_list:
        for point in point_reward_list:
            customer_reward_summary = update_reward_summary_with_new_reward(customer_reward_summary, point.to_reward_summary())
            
    if stamp_reward_list:
        for stamp in stamp_reward_list:
            customer_reward_summary = update_reward_summary_with_new_reward(customer_reward_summary, stamp.to_reward_summary())
            
    customer.reward_summary             = customer_reward_summary
    
    customer.put()    
    
def update_customer_entitled_voucher_summary(customer):
    vouchers_list           = CustomerEntitledVoucher.list_by_customer(customer)
    customer_entitled_voucher_summary    = {}
    
    if vouchers_list:
        for voucher in vouchers_list:
            customer_entitled_voucher_summary = update_customer_entiteld_voucher_summary_with_customer_new_voucher(customer_entitled_voucher_summary, voucher)
                    
    
    customer.entitled_voucher_summary   = customer_entitled_voucher_summary
    
    customer.put()    

        
def update_transaction_all_entitled_reward_summary(transaction_id):
    customer_transaction = CustomerTransaction.get_by_transaction_id(transaction_id)
    if customer_transaction:
        
        logger.debug('customer_transaction=%s', customer_transaction)
        
        point_reward_list       = CustomerPointReward.list_by_transaction_id(transaction_id)
        stamp_reward_list       = CustomerStampReward.list_by_transaction_id(transaction_id)
        vouchers_list           = CustomerEntitledVoucher.list_by_transaction_id(transaction_id)
        prepaid_reward_list     = CustomerPrepaidReward.list_by_transaction_id(transaction_id)
        
        
        logger.debug('prepaid_reward_list count=%d' , len(prepaid_reward_list))
        
        customer_reward_summary             = {}
        customer_entitled_voucher_summary   = {}
        customer_prepaid_summary            = {}
        
        if point_reward_list:
            for p in point_reward_list:
                customer_reward_summary = update_reward_summary_with_new_reward(customer_reward_summary, p.to_reward_summary())
                
        
        if stamp_reward_list:
            for p in stamp_reward_list:
                customer_reward_summary = update_reward_summary_with_new_reward(customer_reward_summary, p.to_reward_summary())
                
        
        if vouchers_list:
            logger.debug('found entitled voucher list')
            entiteld_voucher_brief          = EntitledVoucherSummary()
            
            for customer_voucher in vouchers_list:
                merchant_voucher = customer_voucher.entitled_voucher_entity
                entiteld_voucher_brief.add(merchant_voucher, [customer_voucher])
                
            
            customer_entitled_voucher_summary = entiteld_voucher_brief.entitled_voucher_summary
            logger.debug('customer_entitled_voucher_summary=%s', customer_entitled_voucher_summary)
        
        if prepaid_reward_list:
            for prepaid in prepaid_reward_list:
                customer_prepaid_summary = update_prepaid_summary_with_new_prepaid(customer_prepaid_summary, prepaid.to_prepaid_summary())
        
        customer_transaction.entitled_reward_summary    = customer_reward_summary
        customer_transaction.entitled_voucher_summary   = customer_entitled_voucher_summary
        customer_transaction.entitled_prepaid_summary   = customer_prepaid_summary
        
        customer_transaction.put()
        
        return customer_transaction
        
    else:
        logger.warn('customer_transaction is not found')
        
@model_transactional(desc="giveaway_birthday_reward_to_customer")                
def giveaway_birthday_reward_to_customer(customer, program_configuration, transact_datetime, merchant_acct ):
    
    try:
        this_year               = transact_datetime.year
        program_key             = program_configuration.get('program_key')
        remarks                 = program_configuration.get('remarks')
        program_desc            = program_configuration.get('desc')
        currency_code           = merchant_acct.currency_code
        
        logger.debug('------> going to check for program_key=%s for %s', program_key, customer.name)
        
        is_entitled_before = Customer.check_birthday_reward_have_entitled_before(customer, this_year, program_key)
        logger.debug('------> Customer(%s) entitled before (%s) =%s %s', customer.name, program_desc)
        
        if is_entitled_before == False:
            transaction_details = create_non_sales_system_transaction(customer, transact_datetime=transact_datetime, remarks=remarks)
            
            logger.debug('transaction_details=%s', transaction_details)
               
            currency_config = get_currency_config(currency_code)
            RewardProgramFactory().get_giveaway_based_reward(customer, transaction_details, 
                                                                                program_configuration_list  = [program_configuration], 
                                                                                currency                    = currency_config
                                                                                )
            Customer.update_customer_entitled_birthday_reward_summary(customer, program_key)
        else:
            logger.info('Customer(%s) have entitled birthday reward (%s) on year %d', customer.name, program_desc, this_year)
    except:
        logger.error('Failed to process for customer=%s', customer.name)
        
@model_transactional(desc="giveaway_membership_reward_to_customer")                
def giveaway_membership_reward_to_customer(customer, program_configuration, transact_datetime, merchant_acct ):
    
    try:
        this_year               = transact_datetime.year
        program_key             = program_configuration.get('program_key')
        program_label           = program_configuration.get('label')
        remarks                 = program_configuration.get('remarks')
        program_desc            = program_configuration.get('desc')
        currency_code           = merchant_acct.currency_code
        
        logger.debug('------> going to check for program_key=%s for %s', program_key, customer.name)
        
        is_entitled_before = Customer.check_membership_year_reward_have_entitled_before(customer, this_year, program_key)
        logger.debug('------> Customer(%s) entitled %s', customer.name, program_desc)
        
        if is_entitled_before == False:
            system_remarks = 'Giveaway reward from %s' % program_label
            transaction_details = create_non_sales_system_transaction(customer, remarks=remarks, system_remarks=system_remarks)
            
            logger.debug('transaction_details=%s', transaction_details)
               
            currency_config = get_currency_config(currency_code)
            RewardProgramFactory().get_giveaway_reward(customer, transaction_details, 
                                                                                program_configuration_list  = [program_configuration], 
                                                                                currency                    = currency_config
                                                                                )
            Customer.update_customer_entitled_membership_reward_summary(customer, program_key)
        else:
            logger.info('Customer(%s) have entitled membership reward (%s) on year %d', customer.name, program_desc, this_year)
    except:
        logger.error('Failed to process for customer=%s, due to %s', customer.name, get_tracelog())  
        raise      
        

def revert_redemption(redemption_details, reverted_by, reverted_datetime=None):
    
    if reverted_datetime is None:
        reverted_datetime = datetime.utcnow()
    
    redeem_transaction_id       = redemption_details.transaction_id
    redeemed_summary            = redemption_details.redeemed_summary
    customer                    = redemption_details.redeemed_customer_acct
    
    if program_conf.REWARD_FORMAT_VOUCHER in redeemed_summary.keys():
        #found voucher have been redeem in the transaction
        
        '''
        Example of voucher redeemed_summary
        
        redeemed_summary: {
            voucher: {
                vouchers: {
                    ag50cmV4LWFkbWluLWRldnI1CxIMTWVyY2hhbnRBY2N0GICAgPiW0IcKDAsSD01lcmNoYW50Vm91Y2hlchiAgIC4iqmKCgw: {
                    amount: 1,
                    customer_entitled_vouchers: [
                        {
                            customer_entitled_voucher_key: "ag50cmV4LWFkbWluLWRldnJKCxIEVXNlchiAgIC4y6nFCAwLEghDdXN0b21lchiAgICE3I-JCQwLEhdDdXN0b21lckVudGl0bGVkVm91Y2hlchiAgIDY36afCgw",
                            redeem_code: "pW1QXNSMVxa6"
                        }
                    ],
                    image_url: "https://backofficedev.augmigo.com/static/app/assets/img/voucher/voucher-sample-image.png",
                    label: "Lemon Glass Ginger"
                    }
                }
            }
        },
        
        '''
        
        
        redeemed_vouchers_details = redeemed_summary.get('voucher').get('vouchers')
        for merchant_voucher_key, details in redeemed_vouchers_details.items():
            redeemed_customer_entitled_vouchers_list = details.get('customer_entitled_vouchers')
            customer_transaction_map_by_transaction_id = {}
            
            for redeemed_customer_entitled_voucher_details in redeemed_customer_entitled_vouchers_list:
                redeemed_customer_entitled_voucher = CustomerEntitledVoucher.fetch(redeemed_customer_entitled_voucher_details.get('customer_entitled_voucher_key'))
                if redeemed_customer_entitled_voucher:
                    redeemed_customer_entitled_voucher.revert_from_redemption()
                    redeemed_customer_entitled_voucher_transaction_id = redeemed_customer_entitled_voucher.transaction_id
                    
                    redeemed_customer_transaction = customer_transaction_map_by_transaction_id.get(redeemed_customer_entitled_voucher_transaction_id)
                    
                    if redeemed_customer_transaction is None:
                        redeemed_customer_transaction = CustomerTransaction.get_by_transaction_id(redeemed_customer_entitled_voucher_transaction_id)
                        customer_transaction_map_by_transaction_id[redeemed_customer_entitled_voucher_transaction_id] = redeemed_customer_transaction
                    
                    logger.debug('customer_transation_key=%s', redeemed_customer_transaction.key_in_str)
                    
                    customer_transaction_entitled_voucher_summary = redeemed_customer_transaction.entitled_voucher_summary
                    logger.debug('customer_transaction_entitled_voucher_summary=%s', customer_transaction_entitled_voucher_summary)
                    
                    '''
                    Example of entitled_voucher_summary
                    entitled_voucher_summary: {
                        ag50cmV4LWFkbWluLWRldnI1CxIMTWVyY2hhbnRBY2N0GICAgPiW0IcKDAsSD01lcmNoYW50Vm91Y2hlchiAgIC4iqmKCgw: {
                            amount: 2,
                            redeem_info_list: [
                                {
                                    effective_date: "14-02-2023",
                                    expiry_date: "14-05-2023",
                                    is_redeem: true,
                                    redeem_code: "LheTuNnRr51j",
                                    redeem_transaction_id: "r230214154932958705"
                                },
                                {
                                    effective_date: "14-02-2023",
                                    expiry_date: "14-05-2023",
                                    is_redeem: true,
                                    redeem_code: "pW1QXNSMVxa6",
                                    redeem_transaction_id: "r230214170508014950"
                                }
                            ],
                            voucher_key: "ag50cmV4LWFkbWluLWRldnI1CxIMTWVyY2hhbnRBY2N0GICAgPiW0IcKDAsSD01lcmNoYW50Vm91Y2hlchiAgIC4iqmKCgw"
                        }
                    },
'''                     
                    
                    customer_transaction_entitled_voucher_details = customer_transaction_entitled_voucher_summary.get(merchant_voucher_key)
                    
                    logger.debug('customer_transaction_entitled_voucher_details=%s', customer_transaction_entitled_voucher_details)
                    logger.debug('Going to unmark voucher have been redeemed from customer.entitled_voucher_summary.redeem_info_list')
                    
                    for redeem_info in customer_transaction_entitled_voucher_details.get('redeem_info_list'):
                        if redeem_info.get('redeem_transaction_id')==redeem_transaction_id:
                            del redeem_info['redeem_transaction_id']
                            del redeem_info['is_redeem']
                            
                        
                    logger.debug('entitled_voucher_summary before reverted=%s', customer.entitled_voucher_summary)
                    logger.debug('redeemed entitled_voucher_summary =%s', redeemed_customer_transaction.entitled_voucher_summary)
                    
                    #redeemed_customer_transaction.entitled_voucher_summary = entitled_voucher_summary
                    redeemed_customer_transaction.put()
                    
                    customer_new_entitled_voucher_summary = update_customer_entiteld_voucher_summary_with_customer_new_voucher(customer.entitled_voucher_summary, redeemed_customer_entitled_voucher)
                    
                    #customer = customer_entitled_voucher.entitled_customer_entity
                    
                    customer.entitled_voucher_summary = customer_new_entitled_voucher_summary
                    
                    logger.debug('entitled_voucher_summary before reverted=%s', customer.entitled_voucher_summary)
                    
                    customer.put()
        
    
    if program_conf.REWARD_FORMAT_POINT in redeemed_summary.keys():
        '''
        Example of point redeemed_summary
        redeemed_summary: {
            point: {
                amount: 2,
                customer_point_rewards: [
                    {
                    key: "ag50cmV4LWFkbWluLWRldnJGCxIEVXNlchiAgIC4y6nFCAwLEghDdXN0b21lchiAgICE3I-JCQwLEhNDdXN0b21lclN0YW1wUmV3YXJkGICAgNjfrZAJDA",
                    redeemed_amount: 2
                    }
                ]
            }
        },
        '''
        redeemed_details_list   = redeemed_summary.get(program_conf.REWARD_FORMAT_POINT).get('customer_point_rewards')
        
        for redeemed_details in redeemed_details_list:
            customer_point_reward = CustomerPointReward.fetch(redeemed_details.get('key'))
            redeemed_amount = redeemed_details.get('amount')
            customer_point_reward.update_used_reward_amount( -redeemed_amount )
            customer.reward_summary[program_conf.REWARD_FORMAT_POINT]['amount'] += redeemed_amount
        
        customer.put()
        
    if program_conf.REWARD_FORMAT_STAMP in redeemed_summary.keys():
        '''
        Example of stamp redeemed_summary
        redeemed_summary: {
            stamp: {
                amount: 2,
                customer_stamp_rewards: [
                    {
                    key: "ag50cmV4LWFkbWluLWRldnJGCxIEVXNlchiAgIC4y6nFCAwLEghDdXN0b21lchiAgICE3I-JCQwLEhNDdXN0b21lclN0YW1wUmV3YXJkGICAgNjfrZAJDA",
                    redeemed_amount: 2
                    }
                ]
            }
        },
        '''
        redeemed_details_list   = redeemed_summary.get(program_conf.REWARD_FORMAT_STAMP).get('customer_stamp_rewards')
        
        for redeemed_details in redeemed_details_list:
            customer_stamp_reward = CustomerStampReward.fetch(redeemed_details.get('key'))
            redeemed_amount = redeemed_details.get('amount')
            customer_stamp_reward.update_used_reward_amount( -redeemed_amount )
            
            customer.reward_summary[program_conf.REWARD_FORMAT_STAMP]['amount'] += redeemed_amount
        
        customer.put()
        
    if program_conf.REWARD_FORMAT_PREPAID in redeemed_summary.keys(): 
        #found voucher have been redeem in the transaction
        '''
        Example of prepaid redeemed_summary
        redeemed_summary: {
            prepaid: {
                amount: 25,
                customer_prepaid_reward: [
                {
                    key: "ag50cmV4LWFkbWluLWRldnJICxIEVXNlchiAgIC4y6nFCAwLEghDdXN0b21lchiAgICE3I-JCQwLEhVDdXN0b21lclByZXBhaWRSZXdhcmQYgICAmP-6lAoM",
                    redeemed_amount: 25,
                    }
                ]
            }
        },
        '''
        redeemed_details_list   = redeemed_summary.get(program_conf.REWARD_FORMAT_PREPAID).get('customer_prepaid_rewards')
        
        for redeemed_details in redeemed_details_list:
            customer_prepaid_reward = CustomerPrepaidReward.fetch(redeemed_details.get('key'))
            if customer_prepaid_reward:
                redeemed_amount = redeemed_details.get('amount')
                customer_prepaid_reward.update_used_reward_amount( -redeemed_amount )
                
                customer.prepaid_summary['amount'] += redeemed_amount
        
        customer.put()
        
    redemption_details.revert(reverted_by, reverted_datetime=reverted_datetime)
    
    
    create_merchant_customer_redemption_reverted_upstream_for_merchant(redemption_details)
    
        
                     
        


'''
Created on 7 Jul 2021

@author: jacklok
'''

from flask import Blueprint, request, session 
from flask_restful import abort
import logging
from trexlib.utils.log_util import get_tracelog
from flask_restful import Api
from trexmodel.utils.model.model_util import create_db_client
from flask.json import jsonify
from datetime import datetime
from trexapi.decorators.api_decorators import auth_token_required,\
    outlet_key_required, test_session_expired, device_is_activated
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.customer_models import Customer,\
    CustomerMembership, CustomerTierMembership
from trexmodel.models.datastore.user_models import User
from trexadmin.libs.http import create_rest_message
from trexadmin.libs.http import StatusCode
from trexmodel.models.datastore.merchant_models import Outlet,\
    MerchantAcct, MerchantUser
from trexapi.forms.customer_api_forms import CustomerDetailsNewForm, CustomerDetailsUpdateForm,\
    CustomerSearchForm
from werkzeug.datastructures import ImmutableMultiDict
from trexapi.controllers.user_api_routes import user_details_dict
from trexmodel.models.datastore.transaction_models import CustomerTransaction
from trexapi.utils.api_helpers import get_logged_in_api_username
from trexmodel.models.datastore.redeem_models import CustomerRedemption
from trexmodel.models.datastore.model_decorators import model_transactional
from trexapi.utils.reward_transaction_helper import revert_redemption,\
    revert_transaction
from trexanalytics.bigquery_upstream_data_config import create_merchant_registered_customer_upstream_for_merchant,\
    create_registered_customer_upstream_for_system
from trexmodel.models.datastore.membership_models import MerchantMembership,\
    MerchantTierMembership
from flask_babel import gettext

logger = logging.getLogger('api')


customer_api_bp = Blueprint('customer_api_base_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/customers')

logger = logging.getLogger('debug')

customer_api = Api(customer_api_bp)



@customer_api_bp.route('/register', methods=['POST'])
@auth_token_required
@outlet_key_required
@device_is_activated
def create_customer():
    customer_data_in_json   = request.get_json()
    register_customer_form  = CustomerDetailsNewForm(ImmutableMultiDict(customer_data_in_json))
    
    logger.debug('create_customer: customer_data_in_json=%s', customer_data_in_json)
    
    try:
        if register_customer_form.validate():
            logger.debug('customer registration input is valid')
            db_client = create_db_client(caller_info="create_customer")
            
            is_email_used           = False
            is_mobile_phone_used    = False
            customer_is_exist       = False
            created_customer        = None
            created_user_acct       = None
            
            outlet_key  = request.headers.get('x-outlet-key')
            acct_id     = request.headers.get('x-acct-id')
            
            logger.debug('outlet_key=%s', outlet_key)
            logger.debug('acct_id=%s', acct_id)
            
            with db_client.context():
                merchant_acct   = MerchantAcct.fetch(acct_id)
                outlet          = Outlet.fetch(outlet_key)
                
                email           = customer_data_in_json.get('email')
                mobile_phone    = customer_data_in_json.get('mobile_phone')
                
                birth_date              = customer_data_in_json.get('birth_date')
                            
                if is_not_empty(birth_date):
                    birth_date = datetime.strptime(birth_date, '%d/%m/%Y')
                
                logger.debug('birth_date=%s', birth_date)

                
                logger.debug('email=%s', email)
                logger.debug('mobile_phone=%s', mobile_phone)
                
                if is_not_empty(email):
                    created_user_acct = User.get_by_email(email)
                    
                    #if created_user_acct and created_user_acct.is_email_verified:
                    if created_user_acct:    
                        is_email_used = True
                
                if is_not_empty(mobile_phone):
                    mobile_phone = mobile_phone.replace(" ", "")
                    created_user_acct = User.get_by_mobile_phone(mobile_phone)
                    
                    #if created_user_acct and created_user_acct.is_mobile_phone_verified:
                    if created_user_acct:
                        is_mobile_phone_used = True
                
                
                logger.debug('is_email_used=%s', is_email_used)
                logger.debug('is_mobile_phone_used=%s', is_mobile_phone_used)
                
                if merchant_acct and outlet:
                        
                    logger.debug('merchant_acct.key_in_str=%s', merchant_acct.key_in_str)
                    logger.debug('outlet.merchant_acct_key=%s', outlet.merchant_acct_key)
                    
                    if merchant_acct.key_in_str == outlet.merchant_acct_key: 
                        logger.debug('Valid granted outlet key for merchant acct')
                                                    
                        if is_email_used:
                            created_customer = Customer.get_by_email(email, merchant_acct=merchant_acct)
                        
                        if created_customer is None:    
                            if is_mobile_phone_used:
                                created_customer = Customer.get_by_mobile_phone(mobile_phone, merchant_acct=merchant_acct)
                        
                        if is_email_used or is_mobile_phone_used:
                            if created_customer is None:
                                logger.debug('User account have been created, but customer account is not yet created')
                                
                                created_user_acct = User.update(created_user_acct,
                                                                name                    = customer_data_in_json.get('name'), 
                                                                email                   = customer_data_in_json.get('email'), 
                                                                gender                  = customer_data_in_json.get('gender'),
                                                                birth_date              = birth_date,
                                                                mobile_phone            = mobile_phone, 
                                                                password                = customer_data_in_json.get('password'),
                                                                )
                                
                                created_customer        = Customer.create_from_user(outlet, created_user_acct, customer_data_in_json.get('merchant_reference_code'))
                            
                            else:
                                customer_is_exist = True
                                logger.warn('Customer account using same email or mobile phone have been created')
                        else:
                            created_customer        = Customer.create( 
                                                            outlet                  = outlet, 
                                                            name                    = customer_data_in_json.get('name'), 
                                                            email                   = customer_data_in_json.get('email'), 
                                                            gender                  = customer_data_in_json.get('gender'),
                                                            birth_date              = birth_date,
                                                            mobile_phone            = mobile_phone, 
                                                            merchant_reference_code = customer_data_in_json.get('merchant_reference_code'), 
                                                            password                = customer_data_in_json.get('password'),
                                                            )
                            
                            create_merchant_registered_customer_upstream_for_merchant(created_customer)
                            create_registered_customer_upstream_for_system(created_customer)
                            
                        logger.debug('created_customer=%s', created_customer)
                    else:
                        logger.warn('Invalid granted outlet key')
                else:
                    logger.warn('Invalid granted outlet key')
            
                if customer_is_exist:
                    if is_email_used==True:
                        return create_rest_message('Email have been taken', status_code=StatusCode.BAD_REQUEST)
                    
                    elif is_mobile_phone_used==True:
                        return create_rest_message('Mobile phone have been taken', status_code=StatusCode.BAD_REQUEST)
                    
                
                if created_customer:
                    created_user_acct = created_customer.registered_user_acct
                    response_data = {
                                'customer_key'              : created_customer.key_in_str,
                                'registered_datetime'       : created_customer.registered_datetime.strftime("%d-%m-%Y %H:%M:%S"),
                                'merchant_reference_code'   : created_customer.merchant_reference_code,
                                'reference_code'            : created_customer.reference_code,
                                'merchant_account_key'      : acct_id,
                                'company_name'              : merchant_acct.company_name,
                                'outlet_key'                : outlet_key,  
                                
                                }
                
                    logger.debug('response_data=%s', response_data)
                    
                    response_data.update(user_details_dict(created_user_acct))
                    
                    return create_rest_message(status_code=StatusCode.OK, **response_data)
                    
                else:
                    return create_rest_message(status_code=StatusCode.BAD_REQUEST)
            
        else:
            logger.warn('customer registration input is invalid')
            error_message = register_customer_form.create_rest_return_error_message()
            
            return create_rest_message(error_message, status_code=StatusCode.BAD_REQUEST)
    except:
        logger.error('Fail to register customer due to %s', get_tracelog())
        
        return create_rest_message(status_code=StatusCode.BAD_REQUEST)


@customer_api_bp.route('/reference-code/<ref_code>', methods=['GET'])
@auth_token_required
@device_is_activated
def read_customer(ref_code):
    
    logger.debug('ref_code=%s', ref_code)
    
    if is_not_empty(ref_code):
        db_client = create_db_client(caller_info="read_customer")
        
        acct_id     = request.headers.get('x-acct-id')
        
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(acct_id)
            customer = Customer.get_by_reference_code(ref_code, merchant_acct)
            
        
        if customer:
        
            user_details = customer.registered_user_acct
            customer_details_dict = customer.to_dict(date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S")
            
            customer_details_dict['customer_key']               = customer.key_in_str
            customer_details_dict['is_email_verified']          = user_details.is_email_verified
            customer_details_dict['is_mobile_phone_verified']   = user_details.is_mobile_phone_verified
            
            if 'entitled_voucher_summary' in customer_details_dict:
                customer_details_dict['voucher_summary']            = customer.entitled_voucher_summary
                del customer_details_dict['entitled_voucher_summary']
            
            customer_details_json =  jsonify(customer_details_dict)
            
            
            logger.debug('customer_details_json=%s', customer_details_json)
            
            return customer_details_json
            
        else:
            logger.warn('Customer with reference code (%s) is not found', ref_code)
            return create_rest_message(status_code=StatusCode.BAD_REQUEST)
            
    else:
        return create_rest_message(status_code=StatusCode.BAD_REQUEST)
    
@customer_api_bp.route('/customer-key/<customer_key>', methods=['GET'])
@auth_token_required
@device_is_activated
def read_customer_by_customer_key(customer_key):
    
    logger.debug('customer_key=%s', customer_key)
    
    if is_not_empty(customer_key):
        db_client = create_db_client(caller_info="read_customer_by_customer_key")
        
        with db_client.context():
            customer = Customer.fetch(customer_key)
            
        
        if customer:
            '''
            customer_property_list = ['name', 'email', 'mobile_phone', 'gender', 'birth_date', 'merchant_reference_code']
            
            response_data = {
                        'customer_key'              : customer.key_in_str,
                        'registered_datetime'       : customer.registered_datetime.strftime("%d-%m-%Y %H:%M:%S"),
                        'merchant_reference_code'   : customer.merchant_reference_code,
                        'merchant_account_key'      : customer.registered_merchant_acct_key,
                        'outlet_key'                : customer.registered_outlet_key,  
                        'user_details'              : user_details_dict(user_details_dict), 
                        }
            '''
            
            user_details = customer.registered_user_acct
            customer_details_dict = customer.to_dict(date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S")
            customer_details_dict['customer_key']               = customer.key_in_str
            customer_details_dict['is_email_verified']          = user_details.is_email_verified
            customer_details_dict['is_mobile_phone_verified']   = user_details.is_mobile_phone_verified
            
            if 'entitled_voucher_summary' in customer_details_dict:
                customer_details_dict['voucher_summary']            = customer.entitled_voucher_summary
                del customer_details_dict['entitled_voucher_summary']
            
            return jsonify(customer_details_dict)
            #return jsonify(response_data)
            
        else:
            logger.warn('Customer with customer_key (%s) is not found', customer_key)
            return create_rest_message(status_code=StatusCode.BAD_REQUEST)
            
    else:
        return create_rest_message(status_code=StatusCode.BAD_REQUEST)    


@customer_api_bp.route('/customer-key/<customer_key>', methods=['PUT'])
@auth_token_required
@device_is_activated
def update_customer(customer_key):
    customer_data_in_json   = request.get_json()
    updating_customer_form  = CustomerDetailsUpdateForm(ImmutableMultiDict(customer_data_in_json))
    
    logger.debug('update_customer: customer_data_in_json=%s', customer_data_in_json)
    
    try:
        if updating_customer_form.validate():
            logger.debug('customer update input is valid')
            db_client = create_db_client(caller_info="update_customer")
            
            is_email_used                           = False
            is_mobile_phone_used                    = False
            customer_is_exist                       = False
            updating_customer                       = None
            checking_user_acct                      = None
            is_customer_email_changed               = False
            is_customer_mobile_phone_changed        = False
            is_used_email_same_user_acct            = False
            is_used_mobile_phone_same_user_acct     = False
            
            with db_client.context():
                email               = customer_data_in_json.get('email')
                mobile_phone        = customer_data_in_json.get('mobile_phone')
                
                if mobile_phone:
                    mobile_phone = mobile_phone.replace(" ", "")
                
                updating_customer   = Customer.fetch(customer_key)
                
                if updating_customer:
                    customer_is_exist = True
                    is_customer_email_changed           = updating_customer.email!=email
                    is_customer_mobile_phone_changed    = updating_customer.mobile_phone!=mobile_phone
                
                if is_customer_email_changed:
                    if is_not_empty(email):
                        checking_user_acct = User.get_by_email(email)
                        
                        if checking_user_acct:
                            is_email_used = True
                            is_used_email_same_user_acct = checking_user_acct.key_in_str == updating_customer.registered_user_acct_key
                        
                if is_customer_mobile_phone_changed:
                    if is_not_empty(mobile_phone):
                        checking_user_acct = User.get_by_mobile_phone(mobile_phone)
                    
                        if checking_user_acct:
                            is_mobile_phone_used = True
                            is_used_mobile_phone_same_user_acct = checking_user_acct.key_in_str == updating_customer.registered_user_acct_key
                                
                logger.debug('is_customer_email_changed=%s', is_customer_email_changed)
                logger.debug('is_customer_mobile_phone_changed=%s', is_customer_mobile_phone_changed)
                logger.debug('is_email_used=%s', is_email_used)
                logger.debug('is_mobile_phone_used=%s', is_mobile_phone_used)
                logger.debug('is_used_email_same_user_acct=%s', is_used_email_same_user_acct)
                logger.debug('is_used_mobile_phone_same_user_acct=%s', is_used_mobile_phone_same_user_acct)
                
                if is_email_used==False and is_mobile_phone_used==False:
                    logger.debug('Going to update customer details') 
                    
                    birth_date              = customer_data_in_json.get('birth_date')
                        
                    if is_not_empty(birth_date):
                        birth_date = datetime.strptime(birth_date, '%d/%m/%Y')
                    
                    logger.debug('birth_date=%s', birth_date)
                    
                    Customer.update(customer=updating_customer, 
                                    name                    = customer_data_in_json.get('name'), 
                                    email                   = customer_data_in_json.get('email'), 
                                    gender                  = customer_data_in_json.get('gender'),
                                    birth_date              = birth_date,
                                    mobile_phone            = mobile_phone, 
                                    password                = customer_data_in_json.get('password'),
                                    )
                
                
                if customer_is_exist:
                    if is_email_used==True:
                        return create_rest_message('Email have been taken', status_code=StatusCode.BAD_REQUEST)
                    
                    elif is_mobile_phone_used==True:
                        return create_rest_message('Mobile phone have been taken', status_code=StatusCode.BAD_REQUEST)
                    else:
                        return create_rest_message(status_code=StatusCode.OK)
                else:
                    return create_rest_message(gettext('Customer is not exist'), status_code=StatusCode.BAD_REQUEST)
            
        else:
            logger.warn('customer registration input is invalid')
            error_message = updating_customer_form.create_rest_return_error_message()
            
            return create_rest_message(error_message, status_code=StatusCode.BAD_REQUEST)
    except:
        logger.error('Fail to register customer due to %s', get_tracelog())
        
        return create_rest_message(status_code=StatusCode.BAD_REQUEST)
        
@customer_api_bp.route('/<ref_code>', methods=['DELETE'])
@auth_token_required 
@device_is_activated
def delete_customer(ref_code):
    if is_not_empty(ref_code):
        is_found = False
        db_client = create_db_client(caller_info="read_customer")
        with db_client.context():
            customer = Customer.get_by_reference_code(ref_code) 
            if customer:
                customer.delete()
                is_found = True
        
        if is_found:
            return create_rest_message(status_code=StatusCode.NO_CONTENT)
        else:
            return create_rest_message(gettext('Customer is not exist'), status_code=StatusCode.BAD_REQUEST)
    else:
        return create_rest_message(status_code=StatusCode.BAD_REQUEST)        


@customer_api_bp.route('/register-as-customer', methods=['POST'])
@auth_token_required
@device_is_activated
def register_user_as_customer():
    user_data_in_json           = request.get_json()
    reference_code              = user_data_in_json.get('reference_code')
    
    logger.debug('register_user_as_customer: user_data_in_json=%s', user_data_in_json)
    
    try:
        if is_not_empty(reference_code):
            logger.debug('customer registration input is valid')
            db_client = create_db_client(caller_info="register_user_as_customer")
            
            created_customer        = None
            existing_user_acct      = None
            is_email_used           = False
            is_mobile_phone_used    = False
            merchant_act_key        = None
            
            merchant_acct           = None
            
            outlet_key  = request.headers.get('x-outlet-key')
            acct_id     = request.headers.get('x-acct-id')
            
            logger.debug('register_user_as_customer: outlet_key=%s', outlet_key)
            logger.debug('register_user_as_customer: acct_id=%s', acct_id)
            logger.debug('register_user_as_customer: reference_code=%s', reference_code)
            
            with db_client.context():
                existing_user_acct  = User.get_by_reference_code(reference_code)
                
                logger.debug('register_user_as_customer: existing_user_acct=%s', existing_user_acct)
                
                if existing_user_acct:
                    
                    outlet              = Outlet.fetch(outlet_key)
                        
                    if outlet:
                        merchant_acct       = outlet.merchant_acct_entity
                        merchant_act_key    = outlet.merchant_acct_key  
                        logger.debug('Valid granted outlet key for merchant acct')
                        
                        created_customer = Customer.get_by_reference_code(reference_code, merchant_acct)
                         
                        logger.debug('register_user_as_customer: created_customer=%s', created_customer) 
                         
                        if created_customer is None:
                            
                            
                            email           = existing_user_acct.email
                            mobile_phone    = existing_user_acct.mobile_phone
                            
                            logger.debug('email=%s', email)
                            logger.debug('mobile_phone=%s', mobile_phone)
                            
                            checking_customer = Customer.get_by_email(email, merchant_acct=merchant_acct) 
                            
                            if checking_customer:
                                is_email_used = True
                            else:
                                if is_not_empty(mobile_phone):
                                    checking_customer = Customer.get_by_mobile_phone(mobile_phone, merchant_acct=merchant_acct)
                                    if checking_customer:
                                        is_mobile_phone_used = True
                            
                            logger.debug('is_email_used=%s', is_email_used)
                            logger.debug('is_mobile_phone_used=%s', is_mobile_phone_used)
                            
                            if is_email_used == False and is_mobile_phone_used == False:
                            
                                created_customer = Customer.create_from_user(outlet, existing_user_acct)
                                
                                create_merchant_registered_customer_upstream_for_merchant(created_customer)
                                create_registered_customer_upstream_for_system(created_customer)
                        
                            
                        logger.debug('created_customer=%s', created_customer)
                        
                    else:
                        logger.warn('Invalid granted outlet key or merchant account id')
                
                if created_customer:
                    
                    
                    
                    response_data = {
                                    'customer_key'              : created_customer.key_in_str,
                                    'registered_datetime'       : created_customer.registered_datetime.strftime("%d-%m-%Y %H:%M:%S"),
                                    'reference_code'            : created_customer.reference_code,
                                    'merchant_account_key'      : merchant_act_key,
                                    'company_name'              : merchant_acct.company_name,
                                    'outlet_key'                : outlet_key,  
                                    #'user_details'              : user_details_dict(existing_user_acct),
                                    }
                    
                    response_data.update(user_details_dict(existing_user_acct))
                    
                    logger.debug('register_user_as_customer debug: response_data=%s', response_data)
                    
                    return create_rest_message(status_code=StatusCode.OK, **response_data)
                    
                else:
                    if is_email_used==True:
                        return create_rest_message('Email have been taken', status_code=StatusCode.BAD_REQUEST)
                    
                    elif is_mobile_phone_used==True:
                        return create_rest_message('Mobile phone have been taken', status_code=StatusCode.BAD_REQUEST)
                    else:
                        return create_rest_message('Failed to register customer', status_code=StatusCode.BAD_REQUEST)
            
        else:
            logger.warn('customer registration input is invalid')
            
            return create_rest_message("Missing register customer input data", status_code=StatusCode.BAD_REQUEST)
    except:
        logger.error('Fail to register customer due to %s', get_tracelog())
        
        return create_rest_message(status_code=StatusCode.BAD_REQUEST)

@customer_api_bp.route('/search/<limit>', methods=['POST'])
@auth_token_required
@device_is_activated
#@test_session_expired
def search_customer(limit):
    
    search_member_data_in_json   = request.get_json()
    search_customer_form         = CustomerSearchForm(ImmutableMultiDict(search_member_data_in_json))
    
    logger.debug('search_member_data_in_json=%s', search_member_data_in_json)
    
    db_client = create_db_client(caller_info="search_customer")
    customer_list = []
    
    if search_customer_form.validate():
        name                        = search_customer_form.name.data
        mobile_phone                = search_customer_form.mobile_phone.data
        email                       = search_customer_form.email.data
        reference_code              = search_customer_form.reference_code.data
        merchant_reference_code     = search_customer_form.merchant_reference_code.data
    
        acct_id     = request.headers.get('x-acct-id')
        
        logger.debug('acct_id=%s', acct_id)
        logger.debug('name=%s', name)
        logger.debug('email=%s', email)
        logger.debug('mobile_phone=%s', mobile_phone)
        logger.debug('reference_code=%s', reference_code)
        logger.debug('merchant_reference_code=%s', merchant_reference_code)
        
        limit_int = int(limit)
        
        with db_client.context():
            merchant_acct   = MerchantAcct.fetch(acct_id)
            (search_results, total_count, next_cursor)  = Customer.search_merchant_customer(merchant_acct, 
                                                                                            name                    = name, 
                                                                                            email                   = email, 
                                                                                            mobile_phone            = mobile_phone, 
                                                                                            reference_code          = reference_code,
                                                                                            merchant_reference_code = merchant_reference_code,
                                                                                            limit                   = limit_int,
                                                                                            )
        
            for c in search_results:
                user_details = c.registered_user_acct
                customer_details_dict = c.to_dict(
                                            date_format="%d-%m-%Y", 
                                            datetime_format="%d-%m-%Y %H:%M:%S",
                                            excluded_dict_properties=['registered_merchant_acct'],
                                            )
                customer_details_dict['customer_key']               = c.key_in_str
                customer_details_dict['is_email_verified']          = user_details.is_email_verified
                customer_details_dict['is_mobile_phone_verified']   = user_details.is_mobile_phone_verified
                
                if 'entitled_voucher_summary' in customer_details_dict:
                    customer_details_dict['voucher_summary']            = c.entitled_voucher_summary
                    #del customer_details_dict['entitled_voucher_summary'] 
                
                
                logger.debug('customer_details_dict=%s', customer_details_dict)
                
                customer_list.append(customer_details_dict)
    
        #return create_rest_message(status_code=customer_list.OK, customer_list=customer_list)
        customer_details_list_json =  jsonify(customer_list)
            
            
        logger.debug('customer_details_list_json=%s', customer_details_list_json)
            
        return customer_details_list_json
    else:
        error_message = search_customer_form.create_rest_return_error_message()
        return create_rest_message(error_message, status_code=StatusCode.BAD_REQUEST)
        #return create_rest_message(gettext('Invalid input'), status_code=StatusCode.BAD_REQUEST)
        #return create_rest_message('Invalid input', status_code=StatusCode.BAD_REQUEST)

@customer_api_bp.route('/reference-code/<reference_code>/list-membership', methods=['GET'])
@auth_token_required
#@test_session_expired
def list_customer_membership(reference_code):
    if is_not_empty(reference_code):
        acct_id   = request.headers.get('x-acct-id')
        db_client = create_db_client(caller_info="list_customer_membership")
        customer_membership_final_list = []
        
        logger.debug('reference_code=%s', reference_code)
        
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(acct_id)
            customer = Customer.get_by_reference_code(reference_code, merchant_acct)
            customer_memberhips_list        = CustomerMembership.list_active_by_customer(customer)
            customer_tier_memberhips_list   = CustomerTierMembership.list_active_by_customer(customer)
            
            logger.debug('customer=%s', customer)
            
            logger.debug('customer_memberhips_list=%s', customer_memberhips_list)
            
            if is_not_empty(customer_memberhips_list):
                merchant_memberships_list = MerchantMembership.list_by_merchant_acct(merchant_acct)
                
                for cm in customer_memberhips_list:
                    for mm in merchant_memberships_list:
                        if mm.key_in_str == cm.merchant_membership_key:
                            customer_membership_data = {
                                                            'key'           : cm.key_in_str,
                                                            'label'         : mm.label,
                                                            'entitled_date' : cm.entitled_date.strftime('%d-%m-%Y'),
                                                            'expiry_date'   : cm.expiry_date.strftime('%d-%m-%Y'),
                                                            'is_tier'       : False,
                                                            }
                            
                            if cm.renewed_date is not None:
                                customer_membership_data['renewed_date'] = cm.renewed_date.strftime('%d-%m-%Y'),
                            
                            customer_membership_final_list.append(customer_membership_data)
                            break
                        
            if is_not_empty(customer_tier_memberhips_list):
                merchant_tier_memberships_list = MerchantTierMembership.list_by_merchant_acct(merchant_acct)
                
                for cm in customer_tier_memberhips_list:
                    for mm in merchant_tier_memberships_list:
                        if mm.key_in_str == cm.merchant_tier_membership_key:
                            customer_membership_data = {
                                                            'key'           : cm.key_in_str,
                                                            'label'         : mm.label,
                                                            'entitled_date' : cm.entitled_date.strftime('%d-%m-%Y'),
                                                            'expiry_date'   : cm.expiry_date.strftime('%d-%m-%Y'),
                                                            'is_tier'       : True,
                                                            }
                            
                            customer_membership_final_list.append(customer_membership_data)
                            break            
        
    
        return jsonify(customer_membership_final_list)
    else:
        return create_rest_message(status_code=StatusCode.BAD_REQUEST)    
                


@customer_api_bp.route('/reference-code/<reference_code>/transaction/limit/<limit>', methods=['GET'])
@auth_token_required
#@test_session_expired
def list_customer_transaction(reference_code, limit):
    
    logger.debug('reference_code=%s', reference_code)
    
    if is_not_empty(reference_code):
        limit_int = int(limit, 10)
        acct_id   = request.headers.get('x-acct-id')
        db_client = create_db_client(caller_info="read_customer_sales_transaction")
        transactions_list = []
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(acct_id)
            customer = Customer.get_by_reference_code(reference_code, merchant_acct)
            
        
        if customer:
            dict_properties  = ['transaction_id', 'invoice_id', 'remarks', 'tax_amount', 'transact_amount', 'reward_giveaway_method',
                               'entitled_reward_summary', 'entitled_voucher_summary', 'entitled_prepaid_summary', 
                               #'transact_outlet_details', 
                               'transact_datetime', 'created_datetime',  'transact_outlet_key', 'is_revert', 'reverted_datetime',
                               'transact_by_username', 'is_reward_redeemed', 'is_sales_transaction', 'allow_to_revert',
                               'is_membership_purchase', 'purchased_merchant_membership_key', 'is_membership_renew',
                               ]
            with db_client.context():
                result       = CustomerTransaction.list_customer_transaction(customer, limit=limit_int)
                for r in result:
                    transactions_list.append(r.to_dict(dict_properties=dict_properties,  date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S"))
            
            return jsonify(transactions_list)
            
        else:
            logger.warn('Customer transaction with reference code (%s) is not found', reference_code)
            return create_rest_message(status_code=StatusCode.BAD_REQUEST)
            
    else:
        return create_rest_message(status_code=StatusCode.BAD_REQUEST)
    
@customer_api_bp.route('/reference-code/<reference_code>/transaction/transaction_id/<transaction_id>/revert', methods=['POST'])
@auth_token_required
def revert_customer_transaction(reference_code, transaction_id):
    
    logger.debug('transaction_id=%s', transaction_id)
    
    if is_not_empty(reference_code) and is_not_empty(transaction_id):
        acct_id   = request.headers.get('x-acct-id')
        db_client = create_db_client(caller_info="revert_customer_sales_transaction")
        
        with db_client.context():
            customer_transactionn       = CustomerTransaction.get_by_transaction_id(transaction_id);
        
        if customer_transactionn:
            with db_client.context():
                merchant_username       = get_logged_in_api_username()
                reverted_by             = MerchantUser.get_by_username(merchant_username)
                
                reverted_datetime_utc   = datetime.utcnow()
                __revert_customer_transaction(customer_transactionn, reverted_by, reverted_datetime=reverted_datetime_utc)
            
            return create_rest_message(status_code=StatusCode.OK, reverted_datetime = customer_transactionn.reverted_datetime.strftime('%d-%m-%Y %H:%M:%S'))
        else:    
            return create_rest_message(gettext('Failed to find transaction'), status_code=StatusCode.BAD_REQUEST)
        
        
            
    else:
        return create_rest_message(gettext('Missing reference code or transaction id'), status_code=StatusCode.BAD_REQUEST)    

@model_transactional(desc="revert_customer_transaction")
def __revert_customer_transaction(customer_transction, reverted_by, reverted_datetime):     
    return revert_transaction(customer_transction, reverted_by, reverted_datetime=reverted_datetime)

    
@customer_api_bp.route('/reference-code/<reference_code>/redemption/limit/<limit>', methods=['GET'])
@auth_token_required
#@test_session_expired
def list_customer_redemption(reference_code, limit):
    
    logger.debug('reference_code=%s', reference_code)
    
    if is_not_empty(reference_code):
        limit_int = int(limit, 10)
        acct_id   = request.headers.get('x-acct-id')
        db_client = create_db_client(caller_info="read_customer_sales_transaction")
        redemptions_list = []
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(acct_id)
            customer = Customer.get_by_reference_code(reference_code, merchant_acct)
            
        
        if customer:
            dict_properties  = ['transaction_id', 'invoice_id', 'remarks', 'reward_format', 'redeemed_amount', 'redeemed_summary',
                                   'redeemed_datetime', 'is_revert', 'reverted_datetime',
                                   'redeemed_by_username', 
                                   ]
            with db_client.context():
                result       = CustomerRedemption.list_customer_redemption(customer, limit=limit_int)
                for r in result:
                    redemptions_list.append(r.to_dict(dict_properties=dict_properties,  date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S"))
            
            return jsonify(redemptions_list)
            
        else:
            logger.warn('Customer transaction with reference code (%s) is not found', reference_code)
            return create_rest_message(status_code=StatusCode.BAD_REQUEST)
            
    else:
        return create_rest_message(status_code=StatusCode.BAD_REQUEST)    
    
@customer_api_bp.route('/reference-code/<reference_code>/redemption/transaction_id/<transaction_id>/revert', methods=['POST'])
@auth_token_required
def revert_customer_redemption(reference_code, transaction_id):
    
    logger.debug('transaction_id=%s', transaction_id)
    
    if is_not_empty(reference_code) and is_not_empty(transaction_id):
        acct_id   = request.headers.get('x-acct-id')
        db_client = create_db_client(caller_info="revert_customer_sales_transaction")
        
        with db_client.context():
            customer_redemption    = CustomerRedemption.get_by_transaction_id(transaction_id);
        
        if customer_redemption:
            with db_client.context():
                merchant_username       = get_logged_in_api_username()
                reverted_by             = MerchantUser.get_by_username(merchant_username)
                
                reverted_datetime_utc   = datetime.utcnow()
                __revert_customer_redemption(customer_redemption, reverted_by, reverted_datetime=reverted_datetime_utc)
            
            return create_rest_message(status_code=StatusCode.OK, reverted_datetime = customer_redemption.reverted_datetime.strftime('%d-%m-%Y %H:%M:%S'))
        else:    
            return create_rest_message(gettext('Failed to find transaction'), status_code=StatusCode.BAD_REQUEST)
        
        
            
    else:
        return create_rest_message(gettext('Missing reference code or transaction id'), status_code=StatusCode.BAD_REQUEST)
    
@model_transactional(desc="revert_customer_redemption")
def __revert_customer_redemption(customer_redemption, reverted_by, reverted_datetime=None):     
       
    return revert_redemption(customer_redemption, reverted_by, reverted_datetime=reverted_datetime)
          


'''
Created on 30 Jun 2021

@author: jacklok
'''

from flask import Blueprint, request, Response, session 
from flask_restful import Resource, abort
import logging
from trexlib.utils.log_util import get_tracelog
from flask_httpauth import HTTPBasicAuth
from trexapi import conf
from flask_restful import Api
from trexmodel.utils.model.model_util import create_db_client
from trexmodel.models.datastore.merchant_models import MerchantUser,\
    MerchantAcct
import hashlib
from random import random
from trexlib.utils.string_util import is_not_empty
from flask.json import jsonify
from trexapi import conf as api_conf
from trexapi.decorators.api_decorators import auth_token_required
from trexmodel.models.datastore.loyalty_models import LoyaltyDeviceSetting
from trexmodel.models.datastore.pos_models import POSSetting
from trexapi.controllers.api_routes import APIBaseResource
from trexmodel.models import merchant_helpers
from trexadmin.libs.http import create_rest_message, StatusCode
from trexmodel.models.datastore.customer_models import Customer

#logger = logging.getLogger('api')
logger = logging.getLogger('debug')

auth = HTTPBasicAuth()
#auth = HTTPBasicAuthWrapper()


merchant_api_bp = Blueprint('merchant_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/merchant')

merchant_api = Api(merchant_api_bp)



@merchant_api_bp.route('/ping', methods=['GET'])
def ping():
    return 'Pong', 200

@auth.verify_password
def verify_user_auth(username, password):
    if not (username and password):
        return False
    
    db_client   = create_db_client(caller_info="verify_user_auth")
    valid_auth  = False
    
    logger.debug('username=%s', username)
    logger.debug('password=%s', password)
    
    with db_client.context():
        merchant_user = MerchantUser.get_by_username(username)
        
        logger.debug('merchant_user=%s', merchant_user)
        
        if merchant_user:
            
            md5_hashed_password = hashlib.md5(password.encode('utf-8')).hexdigest()
            
            logger.debug('verify_user_auth: md5_hashed_password=%s', md5_hashed_password)
            
            if merchant_user.is_valid_password(md5_hashed_password):
                valid_auth = True
            else:
                logger.warn('Invalid merchant password')
        else:
            logger.warn('Invalid merchant username=%s', username)    
        
    return valid_auth

@merchant_api_bp.route('/<merchant_acct_code>/details', methods=['GET'])
def read_merchant_acct(merchant_acct_code):
    
    merchant_acct           = None
    if is_not_empty(merchant_acct_code):
        db_client = create_db_client(caller_info="read_merchant_acct")
        with db_client.context():
            merchant_acct = MerchantAcct.get_by_account_code(merchant_acct_code)
            logger.debug('website=%s', merchant_acct.website)
            merchant_info = merchant_helpers.construct_merchant_acct_info(merchant_acct) 
            
        return create_rest_message(status_code=StatusCode.OK,
                                               **merchant_info
                                               )
    else:
        return create_rest_message("Missing merchant account key", status_code=StatusCode.BAD_REQUEST)

@merchant_api_bp.route('/<merchant_key>/customer-details/reference-code/<reference_code>', methods=['GET'])
def read_merchant_customer_details(merchant_key, reference_code):
    
    merchant_acct           = None
    customer_details_dict   = {} 
    
    if is_not_empty(merchant_key) and is_not_empty(reference_code):
        db_client = create_db_client(caller_info="read_merchant_customer_details")
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(merchant_key)
            customer = Customer.get_by_reference_code(reference_code, merchant_acct) 
        
            if customer:
            
                user_details = customer.registered_user_acct
                customer_details_dict = customer.to_dict(date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S")
                customer_details_dict['customer_key']               = customer.key_in_str
                customer_details_dict['is_email_verified']          = user_details.is_email_verified
                customer_details_dict['is_mobile_phone_verified']   = user_details.is_mobile_phone_verified
                '''
                if 'entitled_voucher_summary' in customer_details_dict:
                    customer_details_dict['voucher_summary']            = customer.entitled_voucher_summary
                    del customer_details_dict['entitled_voucher_summary']
                '''
        return jsonify(customer_details_dict)
            
    
    else:
        return create_rest_message("Missing merchant account key or user reference code", status_code=StatusCode.BAD_REQUEST)


class AccountActivatedAPIResource(APIBaseResource):  
    
    @auth.login_required
    def post(self):
        username    = auth.current_user()
        
        acct_id     = request.headers.get('x-acct-id')
        api_key     = request.headers.get('x-api-key')
        outlet_key  = request.headers.get('x-outlet-key')
        
        logger.debug('username=%s', username)
        logger.debug('acct_id=%s', acct_id)    
        logger.debug('api_key=%s', api_key)
        logger.debug('outlet_key=%s', outlet_key)
        
        is_valid            = False
        is_authorized       = False
        output_json = {}
        
        if is_not_empty(acct_id) and is_not_empty(api_key):
            db_client = create_db_client(caller_info="AccountActivatedAPIResource.post")
            with db_client.context():
                merchant_acct = MerchantAcct.fetch(acct_id)
                
                #logger.debug('merchant_acct=%s', merchant_acct)
                
                if merchant_acct:
                    logger.debug('merchant_acct api key=%s', merchant_acct.api_key)
                    
                    if api_key == merchant_acct.api_key:
                        merchant_user               = MerchantUser.get_by_username(username)
                        
                        is_authorized = self.is_outlet_authorized(merchant_user, outlet_key)
                        if is_authorized:
                            (expiry_datetime, token)    = self.generate_token(acct_id, username)
                            logger.debug('outlet is_authorized=%s', is_authorized)
                            logger.debug('expiry_datetime=%s', expiry_datetime)
                            
                            output_json =  {
                                            'auth_token'        : token,
                                            'expires_in'        : int(api_conf.API_TOKEN_EXPIRY_LENGTH_IN_MINUTE) * 60,
                                            #'expiry_datetime'   : expiry_datetime.strftime('%d-%m-%Y %h:%M:$S'),
                                            'granted_outlet'    : merchant_user.granted_outlet_details_list,
                                            'username'          : merchant_user.username,
                                            'name'              : merchant_user.name,
                                            'is_admin'          : merchant_user.is_admin,
                                            'granted_access'    : merchant_user.permission.get('granted_access'),
                                            'gravatar_url'      : merchant_user.gravatar_url,
                                            }
                
                            logger.debug('output_json=%s', output_json)
                            
                            is_valid = True
                    
            if is_valid:
                return output_json
            
            elif is_authorized==False:
                abort(400, msg=["User is not authorized due to outlet is not granted"])
                    
        abort(400, msg=["Failed to authenticate"])
        
    
    def is_outlet_authorized(self, merchant_user, outlet_key):
        return True 
    
class AuthenticateAPIResource(APIBaseResource):  
    
    @auth.login_required
    def post(self):
        username    = auth.current_user()
        
        acct_id     = request.headers.get('x-acct-id')
        api_key     = request.headers.get('x-api-key')
        outlet_key  = request.headers.get('x-outlet-key')
        
        logger.debug('acct_id=%s', acct_id)    
        logger.debug('api_key=%s', api_key)
        logger.debug('outlet_key=%s', outlet_key)
        
        is_authenticated    = False
        is_authorized       = False
        output_json = {}
        
        if is_not_empty(acct_id) and is_not_empty(api_key):
            db_client = create_db_client(caller_info="AuthenticateAPIResource.post")
            with db_client.context():
                merchant_acct = MerchantAcct.fetch(acct_id)
            
                if merchant_acct:
                    if api_key == merchant_acct.api_key:
                        merchant_user               = MerchantUser.get_by_username(username)
                        
                        is_authorized = self.is_outlet_authorized(merchant_user, outlet_key)
                        logger.debug('outlet is_authorized=%s', is_authorized)
                        
                        if is_authorized:
                            (expiry_datetime, token)    = self.generate_token(acct_id, username)
                            #session['auth_username']    = username
                            #session['acct_id']          = acct_id
                            
                            logger.debug('token=%s', token)
                            logger.debug('expiry_datetime=%s', expiry_datetime)
                            logger.debug('auth_username=%s', username)
                
                            output_json =  {
                                            'auth_token'        : token,
                                            'expires_in'        : int(api_conf.API_TOKEN_EXPIRY_LENGTH_IN_MINUTE) * 60,
                                            #'expiry_datetime'   : expiry_datetime.strftime('%d-%m-%Y %h:%M:$S'),
                                            'granted_outlet'    : merchant_user.granted_outlet_details_list,
                                            'username'          : merchant_user.username,
                                            'name'              : merchant_user.name,
                                            'is_admin'          : merchant_user.is_admin,
                                            'granted_access'    : merchant_user.permission.get('granted_access'),
                                            'gravatar_url'      : merchant_user.gravatar_url,
                                            }
                
                            logger.debug('output_json=%s', output_json)
                            
                            is_authenticated = True
                    
            if is_authenticated:
                return output_json
            elif is_authorized==False:
                abort(400, msg=["User is not authorized due to outlet is not granted"])
                    
        abort(400, msg=["Failed to authenticate"])
        
    
    def is_outlet_authorized(self, merchant_user, outlet_key):
        return True    

class ProgramDeviceAuthenticate(AccountActivatedAPIResource):
    
    def is_outlet_authorized(self, merchant_user, outlet_key):
        if merchant_user.is_admin:
            logger.debug('it is merchant admin user, thus it is authorized for all outlet')
            return True
        else:
            data_in_json        = request.get_json()
            activation_code     = data_in_json.get('activation_code')
            device_setting      = LoyaltyDeviceSetting.get_by_activation_code(activation_code)
            assigned_outlet_key = device_setting.assigned_outlet_key
            
            logger.debug('device assigned outlet name=%s', device_setting.assigned_outlet_entity.name)
            
            if outlet_key == device_setting.assigned_outlet_key:
                logger.debug('login outlet is same as device setting assigned outlet')
                if outlet_key in merchant_user.granted_outlet:
                    logger.debug('login outlet is authorized by login merchant user granted outlet')
                    if assigned_outlet_key in merchant_user.granted_outlet:
                        logger.debug('device assigned outlet is authorized by login merchant user granted outlet')
                        return True
                    else:
                        logger.debug('device setting assigned outlet is not granted outlet')
                else:
                    logger.debug('login outlet is not granted outlet')    
            else:
                logger.debug('login outlet is not same as device setting assigned outlet')
                        
        return False
            
class POSDeviceAuthenticate(AccountActivatedAPIResource):
    def is_outlet_authorized(self, merchant_user, outlet_key):
        if merchant_user.is_admin:
            return True
        else:
            data_in_json        = request.get_json()
            activation_code     = data_in_json.get('activation_code')
            device_setting      = POSSetting.get_by_activation_code(activation_code)
            assigned_outlet_key = device_setting.assigned_outlet_key
            
            logger.debug('device assigned outlet name=%s', device_setting.assigned_outlet_entity.name)
            
            if outlet_key == device_setting.assigned_outlet_key:
                logger.debug('login outlet is same as device setting assigned outlet')
                if outlet_key in merchant_user.granted_outlet:
                    logger.debug('login outlet is authorized by login merchant user granted outlet')
                    if assigned_outlet_key in merchant_user.granted_outlet:
                        logger.debug('device assigned outlet is authorized by login merchant user granted outlet')
                        return True
                    else:
                        logger.debug('device setting assigned outlet is not granted outlet')
                else:
                    logger.debug('login outlet is not granted outlet')    
            else:
                logger.debug('login outlet is not same as device setting assigned outlet')
        
        return False            

class SecureAPIResource(AuthenticateAPIResource):  
    
    def __init__(self):
        super(SecureAPIResource, self).__init__()  
    
    
class CheckAuthTokenResource(SecureAPIResource):
    
    @auth_token_required
    def get(self):
        return 'Ping'             
    
     
merchant_api.add_resource(AuthenticateAPIResource,       '/auth')
merchant_api.add_resource(ProgramDeviceAuthenticate,     '/program-auth')
merchant_api.add_resource(POSDeviceAuthenticate,         '/pos-auth')      
    
        
        
        
        


from flask import Blueprint, request, session, jsonify 
from flask_restful import abort
import logging
from trexlib.utils.log_util import get_tracelog
from flask_restful import Api
from trexmodel.utils.model.model_util import create_db_client
#from flask.json import jsonify
from datetime import datetime, timedelta
from trexapi.decorators.api_decorators import auth_token_required,\
    outlet_key_required, user_auth_token_required
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.customer_models import Customer
from trexmodel.models.datastore.user_models import User
from trexadmin.libs.http import create_rest_message
from trexadmin.libs.http import StatusCode
from trexmodel.models.datastore.merchant_models import Outlet,\
    MerchantAcct, MerchantUser
from werkzeug.datastructures import ImmutableMultiDict
from trexmodel.models.datastore.transaction_models import SalesTransaction,\
    CustomerTransaction

from trexapi.utils.reward_transaction_helper import create_sales_transaction,\
    give_reward_from_sales_transaction
from trexapi.utils.api_helpers import get_logged_in_api_username
from trexmodel.models.datastore.voucher_models import MerchantVoucher
from trexapi import conf
from trexmodel.models.datastore.prepaid_models import PrepaidSettings
from trexadmin.libs.jinja.common_filters import format_currency_with_currency_label_filter
from trexadmin.libs.flask.utils.flask_helper import get_merchant_configured_currency_details
from trexmodel import program_conf
from trexapi.forms.sales_api_forms import SalesTransactionForm
from trexlib.utils.crypto_util import encrypt, decrypt, decrypt_json
from trexapi.conf import EARN_INSTANT_REWARD_URL
from trexmodel.models.datastore.model_decorators import model_transactional
from trexanalytics.bigquery_upstream_data_config import create_merchant_customer_transaction_upstream_for_merchant
from flask_babel import gettext



sales_api_bp = Blueprint('sales_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/sales')

logger = logging.getLogger('debug')


@sales_api_bp.route('/ping', methods=['GET'])
def ping():
    return 'pong', 200

@sales_api_bp.route('/create-sales-transaction', methods=['PUT'])
@auth_token_required
@outlet_key_required
def post_sales_transaction():
    
    logger.info('---post_sales_transaction---')
    
    transaction_data_in_json   = request.get_json()
        
    logger.info('transaction_data_in_json=%s', transaction_data_in_json)
    
    transaction_form = SalesTransactionForm(ImmutableMultiDict(transaction_data_in_json))
    
    if transaction_form.validate():
        logger.debug('reward transaction data is valid')
        
        sales_amount        = float(transaction_form.sales_amount.data)
        tax_amount          = transaction_form.tax_amount.data
        invoice_id          = transaction_form.invoice_id.data
        remarks             = transaction_form.remarks.data
        invoice_details     = transaction_data_in_json.get('invoice_details')
        
        transact_datetime   = None
        
        
        if tax_amount is None:
            tax_amount = .0
        else:
            tax_amount = float(tax_amount)
         
        logger.debug('sales_amount=%s', sales_amount)
        logger.debug('tax_amount=%s', tax_amount)
        logger.debug('invoice_id=%s', invoice_id)
        logger.debug('remarks=%s', remarks)
        logger.debug('invoice_details=%s', invoice_details)
        
        db_client = create_db_client(caller_info="post_sales_transaction")
        
        check_transaction_by_invoice_id = None
        
        if is_not_empty(invoice_id):
            with db_client.context():
                check_transaction_by_invoice_id = SalesTransaction.get_by_invoice_id(invoice_id)
        
        if check_transaction_by_invoice_id:
            return create_rest_message("The invoice id have been taken", status_code=StatusCode.BAD_REQUEST)
        else:
            transact_datetime_in_gmt    = transaction_form.transact_datetime.data
            merchant_username           = get_logged_in_api_username()
            
            
            if merchant_username:
                try:
                    with db_client.context():
                        transact_outlet         = Outlet.fetch(request.headers.get('x-outlet-key'))
                        merchant_acct           = transact_outlet.merchant_acct_entity
                        
                            
                        if transact_datetime_in_gmt:
                            transact_datetime    =  transact_datetime_in_gmt - timedelta(hours=merchant_acct.gmt_hour)
                            
                            now                  = datetime.utcnow()
                            if transact_datetime > now:
                                return create_rest_message('Transact datetime cannot be future', status_code=StatusCode.BAD_REQUEST)
                        
                        
                        
                        transact_merchant_user = MerchantUser.get_by_username(merchant_username)
                        
                        sales_transaction = create_sales_transaction( 
                                                                        transact_outlet     = transact_outlet, 
                                                                        sales_amount        = sales_amount,
                                                                        tax_amount          = tax_amount,
                                                                        invoice_id          = invoice_id,
                                                                        remarks             = remarks,
                                                                        transact_by         = transact_merchant_user,
                                                                        transact_datetime   = transact_datetime,
                                                                        invoice_details     = invoice_details,
                                                                    )
                        
                    if sales_transaction:
                        encrypted_transaction_id    = encrypt(sales_transaction.transaction_id)
                        logger.debug('EARN_INSTANT_REWARD_URL=%s', EARN_INSTANT_REWARD_URL)
                        entitled_url                = EARN_INSTANT_REWARD_URL.format(code=encrypted_transaction_id)
                        
                        transaction_details =  {
                                                "entitled_url"      : entitled_url,
                                                "sales_amount"      : str(transaction_form.sales_amount.data),
                                                "invoice_id"        : invoice_id,
                                                }
                            
                    return jsonify(transaction_details)
                except:
                    logger.error('Failed to proceed transaction due to %s', get_tracelog())
                    return create_rest_message('Failed to proceed transaction', status_code=StatusCode.BAD_REQUEST)
                            
            else:
                return create_rest_message('Missing transact user account', status_code=StatusCode.BAD_REQUEST)
        
    else:
        logger.warn('sales transaction data input is invalid')
        error_message = transaction_form.create_rest_return_error_message()
        
        return create_rest_message(error_message, status_code=StatusCode.BAD_REQUEST)

@sales_api_bp.route('/transaction/<encrypted_transaction_id>/read-instant-reward', methods=['GET'])
#@user_auth_token_required
def read_instant_reward_get(encrypted_transaction_id): 
    
    try:
        transaction_id    = decrypt(encrypted_transaction_id)
    except:
        logger.debug('Failed to decrypt transaction id')
        return create_rest_message(gettext('Invalid reward code'), status_code=StatusCode.BAD_REQUEST)    
        
    db_client = create_db_client(caller_info="read_instant_reward_get")
    
    with db_client.context():
        sales_transaction = SalesTransaction.get_by_transaction_id(transaction_id)
    
    if sales_transaction:
        with db_client.context():
            merchant_acct = sales_transaction.transact_merchant_acct
            
        instant_reward = {
                        #"entitled_url"     : EARN_INSTANT_REWARD_URL.format(code=encrypted_transaction_id),
                        'transaction_id'    : transaction_id,
                        'transact_datetime' : sales_transaction.transact_datetime.strftime("%d-%m-%Y %H:%M:%S"),
                        'transact_amount'   : sales_transaction.transact_amount,
                        'used'              : sales_transaction.used,
                        'currency'          : merchant_acct.currency_code,
                        'locale'            : merchant_acct.locale,
                        #'used'             : False,x
                        }
        logger.debug('instant_reward=%s', instant_reward)
        return jsonify(instant_reward) 
    else:
        logger.debug('Invalid transaction id')
        return create_rest_message(gettext('Invalid reward code'), status_code=StatusCode.BAD_REQUEST)
    
    
@sales_api_bp.route('/transaction/<transaction_id>/encrypt-instant-reward-code', methods=['GET'])
#@user_auth_token_required
def encrypt_instant_reward_code_get(transaction_id): 
    
    logger.debug('transaction_id=%s', transaction_id)
    
    encrypted_transaction_id    = encrypt(transaction_id)
    
    logger.debug('encrypted_transaction_id=%s', encrypted_transaction_id)
    
    db_client = create_db_client(caller_info="encrypt_instant_reward_code_get")
    
    with db_client.context():
        sales_transaction = SalesTransaction.get_by_transaction_id(transaction_id)
    
    if sales_transaction.used:
        return create_rest_message(encrypted_transaction_id=encrypted_transaction_id, status_code=StatusCode.BAD_REQUEST)
    else:
        return create_rest_message(encrypted_transaction_id=encrypted_transaction_id, status_code=StatusCode.OK)     
        
    
    
@sales_api_bp.route('/reference-code/<reference_code>/transaction/<encrypted_transaction_id>/customer-earn-instant-reward', methods=['POST'])
@user_auth_token_required
def customer_earn_instant_reward(reference_code, encrypted_transaction_id): 
    
    db_client = create_db_client(caller_info="customer_earn_instant_reward")
        
    logger.info('encrypted_transaction_id=%s', encrypted_transaction_id)
    
    transaction_id = decrypt(encrypted_transaction_id)
    
    logger.info('reference_code=%s', reference_code)
    logger.info('transaction_id=%s', transaction_id)
        
    if is_not_empty(transaction_id):
        with db_client.context():
            sales_transaction = SalesTransaction.get_by_transaction_id(transaction_id)
            
        
        if sales_transaction:
            
            logger.info('sales_transaction is found')
            if sales_transaction.used:
                with db_client.context():
                    customer_transaction = CustomerTransaction.get_by_transaction_id(transaction_id)
                    customer_transaction = customer_transaction.to_dict(date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S")
                
                return jsonify(customer_transaction)
            else:
                try:
                    with db_client.context():
                        merchant_acct = sales_transaction.transact_merchant_acct
                        customer = Customer.get_by_reference_code(reference_code, merchant_acct)
                        
                        if customer is None:
                            logger.info('Customer is not found, thus going to create new customer account')
                            instant_reward_issued_outlet = sales_transaction.transact_outlet_entity
                            
                            reference_code  = request.headers.get('x-reference-code')
                            
                            existing_user_acct = User.get_by_reference_code(reference_code)
                            
                            customer = Customer.create_from_user(instant_reward_issued_outlet, existing_user_acct)
                        
                        if customer:
                            customer_transaction = give_reward_from_sales_transaction(customer, sales_transaction)
                            if customer_transaction:
                                customer_transaction = customer_transaction.to_dict(date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S")
                    
                    if customer:
                        if customer_transaction:
                            return jsonify(customer_transaction)
                        else:
                            return create_rest_message('Failed to earn reward from transaction', status_code=StatusCode.BAD_REQUEST)
                    else:
                        return create_rest_message('Customer Account not found', status_code=StatusCode.BAD_REQUEST)
                except:
                    logger.error('Failed due to %s', get_tracelog())
                    return create_rest_message('Failed to process transaction', status_code=StatusCode.BAD_REQUEST)
            
        else:
            logger.debug('sales_transaction is not found')
            return create_rest_message('Invalid transaction', status_code=StatusCode.BAD_REQUEST)
    else:
        return create_rest_message('Transaction id is empty', status_code=StatusCode.BAD_REQUEST)
            
    
       
